freeze;

/////////////////////////////////////////////////////////////////////////
// baskets.m
/////////////////////////////////////////////////////////////////////////
// Tools for working with baskets of singularities.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////

import "utilities.m": knapsack, toric_knapsack;
import "strings.m": split_to_lc, check_bracket_type, get_brackets, get_seperator;
import "singularities.m": singularity_is_of_type, smooth_in_codim1;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
// possible_baskets
/////////////////////////////////////////////////////////////////////////
// Input: A candidate numerical variety X, and its P_ini contribution.
// Sings is the set of (distinct) singularities we are allowed to use.
// The parameter 'N' can be used to control how many terms of the P_ini
// expansion are used when performing the knapsack.
// Output: A set of all possible baskets (described only by the indices
// of singularities in the seq Sings) that can be made from the given
// singularities, along with a "kernel" (again, given by the indices of
// the singulatities) of collections of singularities whose contribution
// is zero.
/////////////////////////////////////////////////////////////////////////

// Translates e.g. [2,4,1] to [1,1,2,2,2,2,3].
function pttranslate(p)
    return &cat[[i : j in [1..p[i]]] : i in [1..#p]];
end function;

// Rephrases the search as a knapsack problem and returns the result.
function basket_knapsack(Q,w)
    // Is there anything to do?
    if #Q eq 0 then
        return {PowerSequence(Integers())|},{PowerSequence(Integers())|};
    end if;
    // Before using the more powerful toric knapsack machinery, we might
    // be able to use basic linear algebra.
    bool,soln,kernel:=IsConsistent(Matrix(Q),Matrix(1,#w,w));
    if not bool then
        return {PowerSequence(Integers())|},{PowerSequence(Integers())|};
    end if;
    if Dimension(kernel) eq 0 then
        soln:=Eltseq(soln);
        bool,soln:=CanChangeUniverse(soln,Integers());
        if bool and &and[c ge 0 : c in soln] then
            return {PowerSequence(Integers()) | pttranslate(soln)},
                   {PowerSequence(Integers())|};
        else
            return {PowerSequence(Integers())|},{PowerSequence(Integers())|};
        end if;
    end if;
    // The toric knapsack sets up the problem and returns the result in the
    // form of a convex polyhedron, the integral points of which correspond
    // to solutions.
    P:=IntegralPart(toric_knapsack(Q,w));
    points:=[PowerSequence(Integers()) | Eltseq(p) :
                    p in Points(CompactPart(P))];
    relations:=[PowerSequence(Integers()) | Eltseq(p) :
                    p in ZGenerators(InfinitePart(P) : algorithm:="Hemmecke")];
    // These two sets represent the baskets, but not in the right format.
    // We need to return e.g. [2,4,1] as [1,1,2,2,2,2,3] as in pttranslate:
    return {PowerSequence(Integers()) | pttranslate(p) : p in points},
           {PowerSequence(Integers()) | pttranslate(p) : p in relations};
end function;

function possible_baskets(X,Sings,Pini : N:=20)
    // Sanity check
    bool,N:=IsCoercible(Integers(),N);
    error if not bool or N le 1,
        "Parameter 'N' must be a positive integer";
    // Set of baskets (in index form)
    results:={PowerSequence(Integers())|};
    c:=Dimension(X) + CanonicalClass(X) + 1;
    c2:=Floor(c / 2) + 1;
    Xseq:=HilbertCoefficientsInRange(X,c2,c2 + N);
    Piniseq:=CoefficientsInRange(Pini,c2,c2 + N);
    Xdiffseq:=[Universe(Xseq) | Xseq[i] - Piniseq[i] : i in [1..#Xseq]];
    Singseq:=[OrbifoldHilbertCoefficientsInRange(Q,c2,c2 + N) : Q in Sings];
    // Now we find all the ways we can meet P1diff using these
    baskets,kernel:=basket_knapsack(Singseq,Xdiffseq);
    return baskets,kernel;
end function;

/////////////////////////////////////////////////////////////////////////
// estimate_sings
/////////////////////////////////////////////////////////////////////////
// Input: A candidate numerical variety X and singularity types to allow.
// Output: A bool and an indexed set of sings -- the sings are those that
// may appear on X, while the bool is false if there is an easy
// contradiction.
/////////////////////////////////////////////////////////////////////////

// Generate a set of all possible sings (of given singtype) of index r that
// could live on a dim-dimensional variety in P(W). This can contain too many
// sings, as long as their Porb can be calculated, so we essentially ignore the
// singtype. We don't attempt to remove r from W since it might not be there
// (e.g. can have 1/5(1,1,4) on X_30 in P(1,1,4,10,15)).
function make_all_sings(r,W,k,dim,singtype)
    // If necessary, split singtype into a sequence
    if Type(singtype) eq MonStgElt then
        singtype:=split_to_lc(singtype);
    end if;
    // If the singularities should be isolated then we might be able to
    // refine the list of possible weights.
	if "isolated" in singtype then
		W:=[Integers() | a : a in W | GCD(r,a) eq 1];
	end if;
	// Is there anything to do?
	sings:={@ PowerStructure(FmtSng)| @};
	if dim gt #W then return sings; end if;
	// We can only handle quotient singularities
	error if not "quotient" in singtype,
	   "Cannot handle singularities that are not quotient singularities";
    // Seek 1/r(a1,a2,..a_dim) s.t. sum(ai) + k = i*r for some i.
    // (Biggest poss i is (sum(ai) + k) / r, where a1 >= a2, etc.)
    Sort(~W);
    smallest_i:=Ceiling((&+W[1..dim] + k) / r);
    Reversion(~W);
    biggest_i:=Floor((&+W[1..dim] + k) / r);
    for i in [smallest_i..biggest_i] do
        for A in knapsack(W,i * r - k,dim) do
            // Check for (and disallow) stacky behaviour
            if smooth_in_codim1(r,A) then
                P:=Singularity(r,A,k);
                if singularity_is_of_type(P,singtype) then
                    Include(~sings,P);
                end if;
            end if;
        end for;
    end for;
    // Return the singularities
    return sings;
end function;

function estimate_sings(X,singtype)
    // Initialise/collect the data we'll need
    result:={@ PowerStructure(FmtSng)| @};
    W:=Weights(X);
    k:=CanonicalClass(X);
    F:=Format(X);
    // If necessary, split singtype into a sequence
    if Type(singtype) eq MonStgElt then
        singtype:=split_to_lc(singtype);
    end if;
    // The possible indices of singularities are inherited from the ambient wps.
    poss_inds:=[Integers() | r : r in [2..Maximum(W)] |
                      GCD([Integers() | w : w in W | IsDivisibleBy(w,r)]) eq r];
    for r in poss_inds do
        sings_for_r:={@ PowerStructure(FmtSng)| @};
        // There is an infinite set of ambient 1/r points.
        // We handle what we can...
        if "terminal" in singtype and Dimension(X) eq 2 then
            // So X is required to be nonsingular; we literally do nothing.
        elif "logterminal" in singtype and Dimension(X) eq 2 then
            // Looking for 1/r(a,b) with hcf(r,a*b) = 1.
            poss_polns:=[Integers() | w mod r : w in W | GCD(w,r) eq 1];
            for i in [1..#poss_polns - 1] do
                for j in [i + 1..#poss_polns - 1] do
                    a:=poss_polns[i];
                    b:=poss_polns[j];
                    if IsDivisibleBy(a + b + k,r) then
                        Include(~sings_for_r,Singularity(r,[a,b],k));
                    end if;
                end for;
            end for;
        elif "canonical" in singtype and Dimension(X) eq 2 then
            // Looking for 1/r(a,r-a) with hcf(r,a) = 1 AND r dividing k.
            // So search for [a,r-a] in W - but don't check the format to see
            // whether this makes any sense.
            if k mod r ne 0 then
                // The point is not Gorenstein so cannot be a canonical sing;
                // we do nothing.
            else
                // Seek pairs [a,r-a] in W.
                Wmodr:=[Integers() | w mod r : w in W | GCD(w,r) eq 1];
                if r eq 2 then
                    if #Wmodr ge 2 then
                        Include(~sings_for_r,Singularity(2,[1,1],k));
                    end if;
                else
                    for a in {Integers() | a : a in Wmodr |
                                               a lt r / 2 and r - a in Wmodr} do
                        Include(~sings_for_r,Singularity(r,[a,r - a],k));
                    end for;
                end if;
            end if;
        elif "terminal" in singtype and Dimension(X) eq 3 then
            // Search for triples [a,r-a,-k] in W -- but don't consult the
            // format to see if they can be realised. Note that every terminal
            // 3-fold sing includes these 3 wts, even those non-qt sings on
            // Mori's list [see YPG].
            Wmodr:=[Integers() | w mod r : w in W | GCD(w,r) eq 1];
            indrminusk:=Index(Wmodr,(-k) mod r);
            if indrminusk ne 0 then
                // We have got a canonical polarisation, at least
                Remove(~Wmodr,indrminusk);
                if r eq 2 then
                    if #Wmodr ge 2 then // 2 bcs we've already counted -k.
                        Include(~sings_for_r,Singularity(2,[1,1,1],k));
                    end if;
                else
                    // Search through wts < r/2 and see if it works...
                    for a in {Integers() | a : a in Wmodr | a lt r / 2 and
                                                             r - a in Wmodr} do
                        Include(~sings_for_r,Singularity(r,[a,r - a,-k],k));
                    end for;
                end if;
            end if;
        // THINK: could do elif "gorenstein" in singtype for all quotiet/dims.
        else
            // We use the generic code -- we don't have any special tricks!
            sings_for_r:=make_all_sings(r,W,k,#W - 1 - Codimension(F),singtype);
        end if;
        result join:= sings_for_r;
    end for;
    // Return success followed by the result
    return true,result;
end function;

/////////////////////////////////////////////////////////////////////////
// get_baskets
/////////////////////////////////////////////////////////////////////////
// Input: A candidate numerical variety X and singularity types to allow.
// Output: A list of baskets which realise the orbifold Riemann-Roch for
// X. Also gives a generating set for the basket kernel, if any.
/////////////////////////////////////////////////////////////////////////

// Returns true iff the basket B (of singularities Q) along with P_ini P
// generate the Hilbert series h.
function check_hilb(h,P,B,Q)
    return h - P - &+[Parent(h) | OrbifoldHilbertSeries(Q[i]) : i in B ] eq 0;
end function;

function get_baskets(X,singtype)
    // If necessary, split singtype into a sequence
    if Type(singtype) eq MonStgElt then
        singtype:=split_to_lc(singtype);
    end if;
    // Collect the Hilbert series data we need
	h:=HilbertSeries(X);
	Pini:=OrbifoldInitialHilbertSeries(X);
	results:={PowerSequence(PowerStructure(FmtSng))|};
	tailresults:={PowerSequence(PowerStructure(FmtSng))|};
	// Check whether X might have empty basket
	if check_hilb(h,Pini,[],{@ @}) then
		Include(~results,[PowerStructure(FmtSng)|]);
	end if;
    if "nonsingular" in singtype or "smooth" in singtype then
        return results,tailresults;
    end if;
	// Now get the possible singularities...
	bool,Q:=estimate_sings(X,singtype);
	if bool then
		// ...and consider the possible baskets
		Bposs,Btailcone:=possible_baskets(X,Q,Pini);
		for B in Bposs do
			if check_hilb(h,Pini,B,Q) then
				Include(~results,[Q[i] : i in B]);
			end if;
		end for;
		// Also verify that the kernel really contributes nothing
		for B in Btailcone do
			if check_hilb(Parent(h) ! 0,Parent(h) ! 0,B,Q) then
				Include(~tailresults,[Q[i] : i in B]);
			end if;
		end for;
	end if;
	// Return the results
	return results,tailresults;
end function;

/////////////////////////////////////////////////////////////////////////
// Intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic BasketToString( B::SeqEnum[FmtSng] : brackets:="{", latex:=false )
        -> MonStgElt
{A string representation of the basket B of singularities}
    // Sanity check
    bool,msg:=check_bracket_type(brackets);
    require bool: msg;
    require Type(latex) eq BoolElt: "Parameter 'latex' must be a boolean";
    // Get the bracket and seperator
    brackets,left,right:=get_brackets(brackets,latex);
    sep:=get_seperator(",");
    // Is there anything to do?
    if #B eq 0 then
        return brackets select left cat " " cat right else "";
    end if;
    // Construct the string
    BB:=Sort(DistinctSingularities(B));
    s:="";
    times:=latex select "\\times" else " x ";
    l:=latex select "Latex" else "Minimal";
    for Q in BB do
        num := #[i : i in [1..#B] | B[i] eq Q];
        if num eq 1 then
            s cat:= SingularityToString(Q : level:=l) cat sep;
        else
            s cat:= IntegerToString(num) cat times
                        cat SingularityToString(Q : level:=l) cat sep;
        end if;
    end for;
    if brackets then
        return left cat " " cat s[1..#s-2] cat " " cat right;
    else
        return s[1..#s-2];
    end if;
end intrinsic;

intrinsic Baskets(X::FmtCnd) -> SetEnum[FmtSng],SetEnum[SeqEnum[FmtSng]]
{A list of baskets which realise the orbifold Riemann-Roch for the candidate numerical variety X. Also gives a generating set for the basket kernel, if any.}
    if not assigned X`baskets then
        // Fetch the singularity types to allow
        singtype:=split_to_lc(SingularityType(X));
        // Nonsingular case first
        if "nonsingular" in singtype or "smooth" in singtype then
            results,tailcone:=get_baskets(X,singtype);
        // Dimension 2 case
        elif Dimension(X) eq 2 then
            results,tailcone:=get_baskets(X,singtype);
        // Dimension 3 case
        elif Dimension(X) eq 3 then
            if "isolated" in singtype and "quotient" in singtype then
                results,tailcone:=get_baskets(X,singtype);
            else
                require "terminal" in singtype:
                    "The singularity type must be terminal.";
                results,tailcone:=get_baskets(X,singtype);
            end if;
        // General case
        else
            require "isolated" in singtype and "quotient" in singtype:
                "Singularities must be isolated and quotient";
            results,tailcone:=get_baskets(X,singtype);
        end if;
        // Save the results
        X`baskets:=results;
        X`baskernel:=tailcone;
        // Is this candidate reasonable?
        if #results eq 0 then X`reasonable:=false; end if;
    end if;
    if not assigned X`baskernel then
        X`baskernel:={};
    end if;
    // Return the results
    return X`baskets,X`baskernel;
end intrinsic;
