freeze;

/////////////////////////////////////////////////////////////////////////
// candidate.m
/////////////////////////////////////////////////////////////////////////
// The specification, creation and access intrinsics for the FmtCnd
// data type, which is the type of candidate numerical varieties.
// See [BRZ] = [Buckley,Reid,Zhou 2012] for computation of Hilbert
// series and [BK] = [Brown,Kasprzyk 2012] for more description.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////
// A candidate numerical variety is a collection of some of the data
// associated with the graded ring R(Y) of a variety Y in weighted
// projective space (wps). The key piece of data is the Hilbert series of
// R(Y). Typically, we seek varieties Y that arise as a regular pullback
// from the key variety of a format, and so the candidate also records
// the format. (This is back to front thinking: the format comes first,
// and is usually critical to the calculation of the Hilbert series - as
// far as the candidate is concerned, that is the point of the format.)
//
// However, there is no reason why the variety Y should exist. The main
// use of candidate numerical varieties is as collections of data that
// are not yet known to come from some Y, nor even known to be consistent.
// If such a variety does exist, we say that Y REALISES the candidate.
// This package has no code to check whether a given candidate numerical
// variety can be realised or not.
//
// Notation
// --------
// We often denote a candidate numerical variety by X, and even write
// X \subset P(a_1,...,a_n) as though X is a variety - this never implies
// that we know X can be realised unless we say so explicitly.
// We denote the Hilbert series of X by P(X).
// The format is denoted by F; it does correspond to an affine key
// variety V, and the hope is that some regular pullback of V (and
// quotient by C^*) gives a variety that realises X.
//
// If we know a (singularity) basket for X, then we often denote it B.
// (In fact, we consider a sequence of baskets, since in principle X may
// admit different baskets that are consistent with its Hilbert series.
// However in many cases, this sequence has only one element, so we often
// speak as though the basket B is unique.)
/////////////////////////////////////////////////////////////////////////

import "strings.m": num_to_latex, intseq_to_str, split_to_lc, weights_to_string;
import "singularities.m": check_properties, singularity_is_of_type;
import "format.m": max_weight_non_cone;

declare type FmtCnd;

declare attributes FmtCnd:
	format,            // The format of X
	weights,           // The ambient weights a_1..a_n of X
	hilbertseries,     // The Hilbert series P(X) of X
	p_ini,             // The term P_ini in P(X) [see BRZ]
	first20coeffs,     // A sequence of the first 20 coeffs of P(X) when
                       // expanded as a power series (includes const coeff).
	baskets,           // A sequence of possible baskets B of X
	baskernel,         // A set of generators of a cone (whose coordinate
                       // functions index singularities in a list) all of
                       // whose points are baskets which match p_ini
	singtype,          // A comma separated string denoting conditions on
                       // the singularities that may appear in B.
                       // Default is "quotient,isolated" and "terminal"
                       // or "canonical" may be included. (Also
                       // "logterminal", "logcanonical", but they will
                       // build toric varieties to check the results (slow!))
	reasonable,        // True iff there is at least one basket (of the
                       // given singtype) that computes the Hilbert series
	canonicalclass;    // The integer k for which K_X = O(k).

// The candidate cache
cand_cache:=NewStore();

/////////////////////////////////////////////////////////////////////////
// Cache access functions
/////////////////////////////////////////////////////////////////////////

// Clear the cache
procedure cand_cache_clear()
    StoreClear(cand_cache);
end procedure;

// Calculates the P_ini contribution (as a rational function) given the
// numerical data. Assumes that c >= 0.
function calculate_p_ini(n,k,c,c2,Pi)
    // Create the rings
    R:=PolynomialRing(Rationals(),c2);
    S:=PolynomialRing(R);
    t:=S.1;
    // Start calculating
    hilb_to_c2:=S ! 1 + &+[S | Pi[i] * t^i : i in [1..c2]];
    unknown_num_to_c2:=S ! 1 + &+[S | R.i * t^i : i in [1..c2]];
    denom_to_c2:=&+[S | Binomial(n + i,i) * t^i : i in [0..c2]];
    unknown_bit:=unknown_num_to_c2 * denom_to_c2;
    equations_for_unknown_num:=[R | Coefficient(unknown_bit,i) -
                                      Coefficient(hilb_to_c2,i) : i in [1..c2]];
    matrix:=Matrix(c2,c2,[Rationals() | MonomialCoefficient(f,R.i) :
                                 f in equations_for_unknown_num, i in [1..c2]]);
    target:=Matrix(1,c2,[Rationals() | -MonomialCoefficient(f,R ! 1) :
                                               f in equations_for_unknown_num]);
    num_coeffs:=Eltseq(Solution(matrix,target));
    // Assemble the numerator
    polyring:=PolynomialRing(Rationals());
    ChangeUniverse(~num_coeffs,polyring);
    K:=FieldOfFractions(polyring);
    t:=K.1;
    numerator:=(polyring ! 1) +
                   &+[polyring | num_coeffs[i] * t^i : i in [1..c2]] +
                   &+[polyring | num_coeffs[i] * t^(c - i) : i in [1..c2]] +
                   t^c;
    // Eliminate the double counting at t^(c / 2)
    if c2 ge 1 and IsEven(c) then
    	numerator -:= num_coeffs[c2] * t^c2;
    end if;
    // Divide through by the denominator and return
    return numerator / (1 - t)^(n + 1);
end function;

// Returns the P_ini contribution (as a rational function) for the candidate
// numerical variety X. Note: The denominator is (1 - t)^(Dimension(X) + 1).
function p_ini_from_cache(X)
    // Collect the data we need
    n:=Dimension(X);
    k:=CanonicalClass(X);
    c:=n + k + 1;
    // If c < 0 then P_ini = 0 (see [BRZ])
    if c lt 0 then
    	return RationalFunctionField(Rationals()) ! 0;
    end if;
    c2:=Floor(c / 2);
    Pi:=[Integers() | Plurigenus(X,i) : i in [1..c2]];
    // Consult the cache
    key:=[n,k] cat Pi;
    bool,cache:=StoreIsDefined(cand_cache,Name(Format(X)));
    if bool then
        bool,h:=IsDefined(cache,key);
        if bool then return h; end if;
    else
        cache:=AssociativeArray(PowerSequence(Integers()));
    end if;
    // We missed the cache -- calculate P_ini
    h:=calculate_p_ini(n,k,c,c2,Pi);
    // Update the cache -- we take this oportunity to check that it hasn't
    // gotten out of control
    bool,num:=StoreIsDefined(cand_cache,"p_ini_size");
    if not bool then
        num:=0;
    elif num ge 1000 then
        cand_cache_clear();
        num:=0;
        cache:=AssociativeArray(PowerSequence(Integers()));
    end if;
    cache[key]:=h;
    StoreSet(cand_cache,Name(Format(X)),cache);
    StoreSet(cand_cache,"p_ini_size",num + 1);
    return h;
end function;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// Create a new candidate object with the given format, weights, and singularity
// type.
function raw_candidate(F,W,singtype)
    X:=New(FmtCnd);
    X`format:=F;
    X`weights:=W;
    X`reasonable:=true;
    X`singtype:=Type(singtype) eq MonStgElt select singtype
                                              else Join(Setseq(singtype),",");
    return X;
end function;

/////////////////////////////////////////////////////////////////////////
// Functions for the candidate process
/////////////////////////////////////////////////////////////////////////

// Attempts to construct a candidate numerical variety from the given data.
// Returns true followed by the variety X on success, false otherwise.
function cp_is_candidate(F,W,singtype,filter)
    X:=raw_candidate(F,W,singtype);
    if IsReasonable(X) and (filter cmpeq false or filter(X)) then
        return true,X;
    else
        return false,_;
    end if;
end function;

// Returns true iff the process has finished.
function cp_is_empty(info)
    return info[5];
end function;

// Advances the process to the next case.
procedure cp_advance(~info)
    error if info[5], "The process has finished.";
    bool:=false;
    while #info[3] gt 0 and not bool do
        W:=info[3][#info[3]];
        Prune(~info[3]);
        bool,X:=cp_is_candidate(info[1],W,info[4],info[2]);
        if bool then
            info[6] +:= 1;
            info[7]:=X;
        end if;
    end while;
    if not bool then info[5]:=true; end if;
end procedure;

// Returns the current candidate numerical variety from the process.
function cp_current(info)
    error if info[5], "The process has finished.";
    return info[7];
end function;

// Returns the current label from the process.
function cp_label(info)
    error if info[5], "The process has finished.";
    return info[6];
end function;

// Returns a new candidate process for the given format F, dimension d,
// canonical class k, singularity types singtype, minimum weight minwt, and
// optional user filter function filter.
function cp_create(F,d,k,singtype,minwt,filter)
    // Do we need to split the singularity type?
    if Type(singtype) eq MonStgElt then
        singtype:=split_to_lc(singtype);
    end if;
    // Get the possible embedding weights
    maxwt:=max_weight_non_cone(F);
    l:=d + Codimension(F) + 1;
    weights:=CandidateWeights(F,l,k,maxwt : minweight:=minwt);
    weights:=SetToSequence(weights);
    // Create the filter function
    if assigned F`filter then
        g:=filter cmpeq false select F`filter
                                else func<X | F`filter(X) and filter(X)>;
    else
        g:=filter;
    end if;
    // Create the process info
    bool:=false;
    while #weights gt 0 and not bool do
        W:=weights[#weights];
        Prune(~weights);
        bool,X:=cp_is_candidate(F,W,singtype,g);
    end while;
    info:=bool select <F,g,weights,singtype,false,1,X>
                 else <F,g,weights,singtype,true>;
    // Return the new process
    return InternalCreateProcess("Candidate numerical varieties",info,
                                 cp_is_empty,cp_advance,cp_current,cp_label);
end function;

/////////////////////////////////////////////////////////////////////////
// Creation intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Candidate( F::Fmt, W::SeqEnum[RngIntElt], B::SeqEnum[FmtSng] :
    singtype:="isolated,quotient,canonical" ) -> FmtCnd
{Create a candidate numerical variety in format F with weights W and basket B}
    // Sanity check
    require #W - 1 - Codimension(F) ge 0:
        "The number of weights is incompatible with the codimension";
    require Type(singtype) eq MonStgElt:
        "Parameter 'singtype' must be a string";
    singtype:=split_to_lc(singtype);
    bool,msg:=check_properties(singtype);
    require bool: msg;
    // Check that the basket agrees with the singularity types
    if IsNull(B) then B:=[PowerStructure(FmtSng)|]; end if;
    require &and[singularity_is_of_type(Q,singtype) : Q in B]:
        "The basket is incompatible with parameter 'singtype'";
    // Create the object and initialise the data
    X:=raw_candidate(F,W,singtype);
    X`baskets:=[B];
    // Is this candidate reasonable?
    h:=HilbertSeries(X);
    P:=OrbifoldInitialHilbertSeries(X);
    bool:=h - P - &+[Parent(h) | OrbifoldHilbertSeries(Q) : Q in B ] eq 0;
    X`reasonable := bool;
    // Return the candidate
    return X;
end intrinsic

intrinsic CandidatesFromFormat( F::Fmt, d::RngIntElt, k::RngIntElt :
	singtype:="isolated,quotient,canonical", minweight:=1, filter:=false )
    -> SeqEnum[FmtCnd]
{A sequence containing all reasonable candidate numerical varieties in format F of dimension d and canonical class O(k)}
    // Sanity check
    require d ge 0: "The dimension must be a non-negative integer";
    bool,minweight:=IsCoercible(Integers(),minweight);
    require bool and minweight ge 1:
        "Parameter 'minweight' must be a positive integer";
    require Type(singtype) eq MonStgElt:
        "Parameter 'singtype' must be a string";
    require filter cmpeq false or ISA(Type(filter),Program):
        "Parameter 'filter' must be a function";
    singtype:=split_to_lc(singtype);
    bool,msg:=check_properties(singtype);
    require bool: msg;
    // Create the process
    p:=cp_create(F,d,k,singtype,minweight,filter);
    // Extract the results and return
    results:=[PowerStructure(FmtCnd)|];
    while not IsEmpty(p) do
        Append(~results,Current(p));
        Advance(~p);
    end while;
    return results;
end intrinsic;

intrinsic CandidateProcessFromFormat( F::Fmt, d::RngIntElt, k::RngIntElt :
    singtype:="isolated,quotient,canonical", minweight:=1, filter:=false )
    -> Process
{A process generating all reasonable candidate numerical varieties in format F of dimension d and canonical class O(k)}
    // Sanity check
    require d ge 0: "The dimension must be a non-negative integer";
    bool,minweight:=IsCoercible(Integers(),minweight);
    require bool and minweight ge 1:
        "Parameter 'minweight' must be a positive integer";
    require Type(singtype) eq MonStgElt:
        "Parameter 'singtype' must be a string";
    require filter cmpeq false or ISA(Type(filter),Program):
        "Parameter 'filter' must be a function";
    singtype:=split_to_lc(singtype);
    bool,msg:=check_properties(singtype);
    require bool: msg;
    // Return the process
    return cp_create(F,d,k,singtype,minweight,filter);
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// Display intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Print(X::FmtCnd,l::MonStgElt)
{Print the candidate numerical variety X at print level l}
    if l ne "Minimal" and l ne "Default" and l ne "Maximal" and l ne "Magma" and
       l ne "Latex" then l:="Default"; end if;
    printf Trim(CandidateToString(X : level:=l));
end intrinsic;

intrinsic CandidateToString(X::FmtCnd : level:="Default") -> MonStgElt
{A string representation of the candidate numerical variety X}
    // Sanity check
    require Type(level) eq MonStgElt and (level eq "Minimal" or
        level eq "Default" or level eq "Maximal" or level eq "Magma" or
        level eq "Latex"):
        "Parameter 'level' should be one of \"Minimal\", \"Default\", \"Maximal\", \"Magma\", or \"Latex\"";
    // Is this the LaTeX print level?
    if level eq "Latex" then
        // Create a LaTeX representation of this candidate
        str:=FormatToString(Format(X) : level:="Latex");
        try
            reasonable:=IsReasonable(X);
        catch e;
            reasonable:=false;
        end try;
        if reasonable then
            baskets:=Baskets(X);
            bstr:="";
            for B in baskets do
                bstr cat:= "$" cat BasketToString(B : brackets:="",latex:=true)
                                                                       cat "$&";
            end for;
            str cat:= bstr[1..#bstr - 1] cat "\\\\\n";
            if #baskets eq 1 then
                str cat:= "$" cat BasketToString(Representative(baskets) : brackets:="",
                                                        latex:=true) cat "$&";
            elif #baskets gt 1 then
                bask:="";
                for B in baskets do
                    bask cat:=BasketToString(B : brackets:="", latex:=true)
                                                                       cat ", ";
                end for;
                s cat:= "$" cat bask[1..#bask - 2] cat "$&";
            end if;
            if #baskets eq 1 and Dimension(X) eq 2 and
                CanonicalClass(X) gt 0 then
                ktop,chi:=Explode(PlurigenusInvariants(X));
                str cat:= Sprintf("$%o$&$%o$\\\\\n",num_to_latex(ktop),
                                                             num_to_latex(chi));
            elif #baskets eq 1 and Dimension(X) eq 3 and
                CanonicalClass(X) gt 0 then
                ktop,chi,kc2:=Explode(PlurigenusInvariants(X));
                str cat:= Sprintf("$%o$&$%o$&$%o$\\\\\n",num_to_latex(ktop),
                                           num_to_latex(chi),num_to_latex(kc2));
            end if;
        end if;
        // Return the string
        return str;
    end if;
    // Start creating the strings that we'll need
    if Dimension(X) eq 2 then
        dstring:="surface";
    else
        dstring:=IntegerToString(Dimension(X)) cat "-fold";
    end if;
    try
        wstring:=weights_to_string(Weights(X) : brackets:="(");
        str:=Sprintf("Candidate %o X in P%o with format %o.\n",
                                               dstring,wstring,Name(Format(X)));
    catch e;
        str:=Sprintf("Candidate %o X with format %o.\n",dstring,
                                                               Name(Format(X)));
    end try;
    // If this isn't the minimal print level then we extend
    if level ne "Minimal" then
        try
            canclass:=CanonicalClass(X);
            str cat:= Sprintf("Canonical class: %o\n",canclass);
        catch e;
        end try;
        try
            eqwts:=EquationDegrees(X);
            if #eqwts ne 0 then
                str cat:= Sprintf("Equation degrees: %o\n",
                                                          intseq_to_str(eqwts));
            end if;
        catch e;
        end try;
        if level eq "Maximal" then
            str cat:= Sprintf("Codimension: %o\n",Codimension(X));
        end if;
        try
            reasonable:=IsReasonable(X);
            bool:=true;
        catch e;
            bool:=false;
        end try;
        if bool then
            if reasonable then
                baskets:=Baskets(X);
                if #baskets eq 0 then
                    str cat:= "Basket: { }\n";
                elif #baskets eq 1 then
                    str cat:= "Basket: " cat
                              BasketToString(Representative(baskets)) cat "\n";
                    if level eq "Maximal" then
                        if Dimension(X) eq 2 and CanonicalClass(X) gt 0 then
                            ktop,chi:= Explode(PlurigenusInvariants(X));
                            str cat:= Sprintf("K^2: %o\nchi: %o\n", ktop,chi);
                        elif Dimension(X) eq 3 and CanonicalClass(X) gt 0 and
                                &and[&and[IsTerminalSingularity(p) : p in B] :
                                                              B in baskets] then
                            ktop,chi,kc2:= Explode(PlurigenusInvariants(X));
                            str cat:= Sprintf("K^3: %o\nchi: %o\nK*c_2: %o\n",
                                                                  ktop,chi,kc2);
                        end if;
                    end if;
                elif level eq "Maximal" then
                    bas:="";
                    for B in baskets do
                        bas cat:= BasketToString(B) cat ", ";
                    end for;
                    str cat:= "Baskets: " cat bas[1..#bas - 2] cat "\n";
                end if;
                if assigned X`baskernel and #X`baskernel ne 0 and
                    (#baskets le 1 or level eq "Maximal") then
                    ker:="";
                    for B in X`baskernel do
                        ker cat:= BasketToString(B) cat ", ";
                    end for;
                    str cat:= "Basket kernel: " cat ker[1..#ker - 2] cat "\n";
                end if;
            else
                str cat:= "This numerical variety is not reasonable.\n";
            end if;
        end if;
    end if;
    // Return the string
    return str;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// Access intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Format(X::FmtCnd) -> Fmt
{The underlying format of the candidate numerical variety}
    return X`format;
end intrinsic;

intrinsic IsReasonable(X::FmtCnd) -> BoolElt
{False if the candidate numerical variety X has failed some test in its analysis so far}
    if X`reasonable and not assigned X`baskets then
        _:=Baskets(X);
    end if;
    return X`reasonable;
end intrinsic;

intrinsic SingularityType(X::FmtCnd) -> MonStgElt
{A string describing any restrictions imposed on the singularity types for the candidate numerical variety X}
    return X`singtype;
end intrinsic;

intrinsic ChangeSingularityType(~X::FmtCnd,S::MonStgElt)
{Change the singularity types for the candidate numerical variety X to the string S; any singularity data computed for X so far will be deleted.}
    // Sanity check
    S:=split_to_lc(S);
    bool,msg:=check_properties(S);
    require bool: msg;
    // Change the type
	if assigned X`singtype then
		X`reasonable:=true;
		delete X`baskets;
	end if;
    X`singtype:=Join(Setseq(S),",");
end intrinsic;

intrinsic BettiData(X::FmtCnd) -> SeqEnum
{The Betti data of the candidate numerical variety X}
    return BettiData(Format(X));
end intrinsic;

intrinsic Dimension(X::FmtCnd) -> RngIntElt
{The dimension of the candidate numerical variety X}
    require assigned X`weights: "Weights not assigned";
    return Dimension(Format(X),Weights(X));
end intrinsic;

intrinsic Codimension(X::FmtCnd) -> RngIntElt
{The generic codimension of the candidate numerical variety X}
    return Codimension(Format(X));
end intrinsic;

intrinsic GenericWeights(X::FmtCnd) -> SeqEnum
{The generic weights of the candidate numerical variety X}
    return GenericWeights(Format(X));
end intrinsic;

intrinsic Equations(X::FmtCnd) -> SeqEnum[RngMPolElt]
{The generic equations of the candidate numerical variety X}
    return Equations(Format(X));
end intrinsic;

intrinsic EquationDegrees(X::FmtCnd) -> SeqEnum
{The generic degrees of the equations of the candidate numerical variety X (up to possible masking)}
    return EquationDegrees(Format(X));
end intrinsic;

intrinsic SyzygyDegrees(X::FmtCnd,n::RngIntElt) -> SeqEnum
{The degrees of the n-th syzygies of the candidate numerical variety X}
    return SyzygyDegrees(Format(X),n);
end intrinsic;

intrinsic AdjunctionDegree(X::FmtCnd) -> RngIntElt
{The adjunction degree of the candidate numerical variety X (if Gorenstein)}
    return AdjunctionDegree(Format(X));
end intrinsic;

intrinsic CanonicalClass(X::FmtCnd) -> RngIntElt
{The canonical degree of the candidate numerical variety X (if Gorenstein)}
    if not assigned X`canonicalclass then
        require assigned X`weights: "Weights not assigned";
        X`canonicalclass:=Integers() ! (-&+Weights(X) +
                                            AdjunctionDegree(Format(X)));
	end if;
    return X`canonicalclass;
end intrinsic;

intrinsic Weights(X::FmtCnd) -> SeqEnum[RngIntElt]
{The embedding weights of the candidate numerical variety X}
    require assigned X`weights: "Weights not assigned";
    return X`weights;
end intrinsic;

intrinsic Ambient(X::FmtCnd) -> Prj
{The embedding projective space of the candidate numerical variety X}
    require assigned X`weights: "Weights not assigned";
    return ProjectiveSpace(Rationals(),Weights(X));
end intrinsic;

intrinsic HilbertNumerator(X::FmtCnd) -> RngUPolElt
{The Hilbert numerator of candidate numerical variety X}
    return HilbertNumerator(Format(X));
end intrinsic;

intrinsic HilbertSeries(X::FmtCnd) -> FldFunRatUElt
{The Hilbert series of candidate numerical variety X}
    if not assigned X`hilbertseries then
        require assigned X`weights: "Weights not assigned";
        W:=Weights(X);
        hnum:=HilbertNumerator(Format(X));
        R:=Parent(hnum);
        h:=hnum / &*[R | 1 - R.1^a : a in W];
        X`hilbertseries:=h;
	end if;
    return X`hilbertseries;
end intrinsic;

intrinsic OrbifoldInitialHilbertSeries(X::FmtCnd) -> FldFunRatUElt
{The initial Hilbert series in the orbifold Reimann-Roch of candidate numerical variety X}
    if not assigned X`p_ini then
        X`p_ini:=p_ini_from_cache(X);
    end if;
    return X`p_ini;
end intrinsic;

intrinsic HilbertCoefficients(X::FmtCnd,n::RngIntElt) -> SeqEnum
{The first n coefficients of the Hilbert series of the candidate numerical variety X; that is, the coefficients of 1,t,...,t^(n-1)}
    // Sanity check
    require n ge 0: "Argument 2 must be non-negative";
    // Can we make use of the cache?
    if n le 20 then
    	if not assigned X`first20coeffs then
            h:=HilbertSeries(X);
            X`first20coeffs:=FirstNCoefficients(h,20);
        end if;
        return X`first20coeffs[1..n];
    end if;
    // No -- do the computation
    h:=HilbertSeries(X);
    coeffs:=FirstNCoefficients(h,n);
    // Set the cache (if not set) and return
    if not assigned X`first20coeffs then
        X`first20coeffs:=coeffs[1..20];
    end if;
    return coeffs;
end intrinsic;

intrinsic HilbertCoefficientsInRange(X::FmtCnd,n1::RngIntElt,n2::RngIntElt)
    -> SeqEnum
{The coefficients of t^n1,...,t^n2 of the Hilbert series of the candidate numerical variety X}
    // Sanity check
    require n1 le n2: "Argument 2 must be less than or equal to argument 3";
    // If n2 < 0, then we already know the answer
    if n2 lt 0 then
        return ZeroSequence(Integers(),n2 - n1 + 1);
    end if;
    // If n1 < 0, then we know that the first -n1 coeffs are zero
    initial_zeros:=ZeroSequence(Integers(),Maximum(0,-n1));
    n1:=Maximum(n1,0);
    coeffs:=HilbertCoefficients(X,n2 + 1);
    return initial_zeros cat coeffs[n1 + 1..n2 + 1];
end intrinsic;

intrinsic Plurigenus(X::FmtCnd,n::RngIntElt) -> RngIntElt
{dim H^0(X,nA) where A is the polarisation on the candidate numerical variety X}
    // Sanity check
    require n ge 0: "Argument 2 must be non-negative";
    // Does it make sense to use the cache?
    if n lt 20 then
    	coeffs:=HilbertCoefficients(X,n + 1);
    	if #coeffs ge n + 1 then
    	    return coeffs[n + 1];
    	end if;
    end if;
    // No -- do the computation
    S:=PowerSeriesRing(Rationals(),n + 1);
    return Coefficient(S ! HilbertSeries(X),n);
end intrinsic;
