freeze;

/////////////////////////////////////////////////////////////////////////
// plurigenus.m
/////////////////////////////////////////////////////////////////////////
// Compute the invariants K^n, chi(O_X) and K.c_2(T_X) that occur in
// the plurigenus formula for surfaces and 3-folds.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////

declare attributes FmtCnd:
    ktop,              // The intersection number K_X^n (n = dim X)
    chi,               // The Euler characteristic chi(O_X)
    kc2;               // The intersection number K_X.c_2(T_X)

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// The coefficient of k^3.
function coeffk3(n)
    return (2 * n - 1) * (n - 1) * n / 12;
end function;

// The coefficient of \chi.
function coeffchi(n)
    return 1 - 2 * n;
end function;

// Returns c mod r.
function modr(c,r)
    return c mod r;
end function;

// The contribution from the terminal singularity Q.
function contQ(Q,n,k)	// THINK: k is not used? it's just a summation var below
    r:=Index(Q);
    local_k:=Polarisation(Q);
    wts:=Weights(Q);
    ind:=Index(wts,(r - local_k) mod r);
    Remove(~wts,ind);
    a:=wts[1];
    b:=Integers() ! (Integers(r) ! (local_k / a));
    partII:=&+[Rationals() | modr(b * k,r) * (r - modr(b * k,r)) / (2 * r) :
                                                       k in [1..modr(n,r) - 1]];
    return &+[Rationals() | modr(b * k,r) * (r - modr(b * k,r)) / (2 * r) :
                                                               k in [1..n - 1]];
end function;

// The total contribution from the basket B of terminal singularities.
function contB(B,n,k)
    return &+[Rationals() | contQ(Q,n,k) : Q in B];
end function;

/////////////////////////////////////////////////////////////////////////
// Intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic PlurigenusInvariants(X::FmtCnd) -> SeqEnum[FldRatElt]
{A sequence containing the numbers K_X^2, chi(O(X)) (resp. K_X^3, chi(O(X)) and K_X*c_2(X)) appearing in the plurigenus formula of a 2-dimensional (resp. 3-dimensional) candidate numerical variety X with canonical singularities}
    // Sanity check
    d:=Dimension(X);
    require d eq 2 or d eq 3:
        "Can only compute plurigenus invariants for surfaces and 3-folds";
    // Calculate the surface case
    if Dimension(X) eq 2 then
        if not assigned X`ktop or not assigned X`chi then
            B:=Baskets(X);
            require #B eq 1:
                "There is more than one basket: You must specify which basket to use.";
            ktop,chi:=Explode(PlurigenusInvariants(X,Representative(B)));
            X`ktop:=ktop;
            X`chi:=chi;
        end if;
        return [Rationals() | X`ktop,X`chi];
    // Calculate the 3-fold case
    elif Dimension(X) eq 3 then
        if not assigned X`ktop or not assigned X`chi or not assigned X`kc2 then
            B:=Baskets(X);
            require #B eq 1:
                "There is more than one basket: You must specify which basket to use.";
            ktop,chi,kc2:=Explode(PlurigenusInvariants(X,Representative(B)));
            X`ktop:=ktop;
            X`chi:=chi;
            X`kc2:=kc2;
        end if;
        return [Rationals() | X`ktop,X`chi,X`kc2];
    end if;
end intrinsic;

intrinsic PlurigenusInvariants(X::FmtCnd,B::SeqEnum) -> SeqEnum[FldRatElt]
{A sequence containing the numbers K_X^2, chi(O(X)) (resp. K_X^3, chi(O(X)) and K_X*c_2(X)) appearing in the plurigenus formula of a 2-dimensional (resp. 3-dimensional) candidate numerical variety X with canonical singularities with basket B}
    // Sanity check
    d:=Dimension(X);
    require d eq 2 or d eq 3:
        "Can only compute plurigenus invariants for surfaces and 3-folds";
    // Calculate the surface case
    if Dimension(X) eq 2 then
        k:=CanonicalClass(X);
        // We assume k > 0 for now
        assert k gt 0;
        row1:=[Rationals()|2 * (2 - 1) / 2, 1];
        row2:=[Rationals()|3 * (3 - 1) / 2, 1];  // to compute p2 and p3 by RR.
        coeff_matrix:=Transpose(Matrix(2,2, row1 cat row2));
        p2:=Plurigenus(X,2 * k);
        p3:=Plurigenus(X,3 * k);
        targ_mx:=Matrix(1,2,[p2,p3]);
        first_sol,ker:=Solution(coeff_matrix,targ_mx);
        first_sol:=Eltseq(first_sol);
        assert Dimension(ker) eq 0;     // true: mx = [1,3] [1,1]
        ktop:=first_sol[1];
        chi:=first_sol[2];
        return [Rationals() | ktop,chi];
    // Calculate the 3-fold case
    elif Dimension(X) eq 3 then
        k:=CanonicalClass(X);   // THINK: What if k = 0? errors in contB if k<0
        row1:=[Rationals()|coeffk3(Sign(k) * 2), coeffchi(Sign(k) * 2), contB(B,2,k)];
        row2:=[Rationals()|coeffk3(Sign(k) * 3), coeffchi(Sign(k) * 3), contB(B,3,k)];
        coeff_matrix:=Transpose(Matrix(2,3, row1 cat row2));
        p2:=Plurigenus(X,2 * AbsoluteValue(k));
        p3:=Plurigenus(X,3 * AbsoluteValue(k));
        targ_mx:=Matrix(1,2,[p2,p3]);
        first_sol,ker:=Solution(coeff_matrix,targ_mx);
        assert Dimension(ker) eq 1;	// true: top of mx = [1/2,-3] [5/2,-5]
        ker:=Eltseq(ker.1);
        first_sol:=Eltseq(first_sol);
        assert ker[3] ne 0;     // Else we need more to solve for K^3,chi
        // But also obviously true by same top of mx.
        factor:=(1 - first_sol[3]) / ker[3];
        k3chi:=[Rationals()|factor * ker[i] + first_sol[i] : i in [1..2]];
        ktop:=k3chi[1];
        chi:=k3chi[2];
        // Compute K_X * c_2(X)
        kc2:=&+[Rationals() | Index(Q) - 1 / Index(Q) : Q in B] - 24 * chi;
        return [Rationals() | ktop, chi, kc2];
    end if;
end intrinsic;
