freeze;

/////////////////////////////////////////////////////////////////////////
// singularities.m
/////////////////////////////////////////////////////////////////////////
// A collection of intrinsics for creating and accessing singularity
// data.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////
// Overview
// --------
// We use standard notation 1/r(a_1,a_2,...,a_n) for an n-dimensional
// quotient singularity A^n / (Z/r) acting diagonally via
// diag(eps^a_1,...), where eps^r = 1.
//
// Any varieties X eventually described by this code will be polarised
// globally by an ample divisor A. Locally at each singularity A restricts
// to a divisor in one of the eigenspaces of the Z/r action. This is the
// polarisation of the singularity. Conversely, each singularity can be
// given a polarisation as part of its defining data, and it may only
// appear on varieties whose global polarisation restricts to the given
// local one.
//
// Current Restrictions
// --------------------
// We apply Theorem 1.1 of Buckley-Reid-Zhou (2012) to compute
// contributions to Hilbert series of ISOLATED QUOTIENT singularities,
// and so this version of the code imposes those conditions on
// singularities: the 'isolated' condition is simply that each a_i is
// coprime to r.
//
// Despite these restrictions, the code is prepared so it can admit more
// general singularity types. In particular, as long as the attribute
// 'porb' is set (which could be done by a knowledgeable user), the code
// could work (although any additional requirements on singularity type
// could be checked).
//
// The zero locus of equations inside such quotient singularities define
// more complicated singularities: e.g. the hyperquotient singularity
// 1/r(a_1,a_2,...,a_n; d), which denotes a germ of the origin in (f = 0)
// in 1/r(a_1,a_2,...,a_n) for a polynomial f homogeneous of degree d for
// the Z/r action. While quotient singularities are determined by their
// notation, these more complicated singularities are not (exactly what
// f is matters), so we can rarely state precise properties of the type
// alone, such as whether it is terminal or not.
/////////////////////////////////////////////////////////////////////////

import "strings.m": intseq_to_str, split_to_lc;

declare type FmtSng;

declare attributes FmtSng:
    id,                 // The internal id of this singularity
    hash,               // The hash value
    index,              // The Cartier index of this singularity (usually r)
    weights,            // The defining weights a1,a2,...an of this singularity
    eqn_weights,        // The (local Z/r) degrees of equations of singularity
    polarisation,       // Local polarisation: an integer in range 0,1,..r-1.
    dimension,          // The dimension of the singularity (as a germ)
    porb,               // The orbifold Riemann-Roch contribution P_orb(sing)
    porb_num,           // The natural numerator P_num(sing), defined by
                        //     P_orb = P_num / (1-t)^(dimension(sing) + 1)
    first20coeffs,      // A cache of the first 20 Hilbert coefficients
    is_terminal,        // Is this terminal?
    is_canonical,       // Is this canonical?
    is_logterminal,     // Is this log-terminal?
    is_logcanonical;    // Is this log-canonical?

// The singularity cache
sing_cache:=NewStore();

/////////////////////////////////////////////////////////////////////////
// Cache access functions
/////////////////////////////////////////////////////////////////////////

// Clear the cache
procedure sing_cache_clear()
    StoreClear(sing_cache);
end procedure;

// Returns the d-dimensional singularity 1/r(w) of index r with polarisation k
// and weights w from the cache, or creates it if it doesn't already exist.
function get_sing_from_cache(r,w,k,d,B)
    // Sanity check
    bool,r:=IsCoercible(Integers(),r);
    error if not bool or r le 0, "Argument 1 must be a positive integer";
    if IsNull(w) then
        w:=[Integers()|];
    else
        bool,w:=CanChangeUniverse(w,Integers());
        error if not bool, "Argument 2 must be a sequence of integers";
    end if;
    bool,k:=IsCoercible(Integers(),k);
    error if not bool, "Argument 3 must be an integer";
    bool,d:=IsCoercible(Integers(),d);
    error if not bool or d le 0, "Argument 4 must be a positive integer";
    if IsNull(B) then
        B:=[Integers()|];
    else
        bool,B:=CanChangeUniverse(B,Integers());
        error if not bool, "Argument 5 must be a sequence of integers";
    end if;
    // Normalise the input data
    w:=Sort([Integers() | a mod r : a in w]);
    B:=Sort([Integers() | a mod r : a in B]);
    // Create the key for this data
    key:=intseq_to_str(w : sep:=",") cat "|" cat IntegerToString(r) cat "|"
              cat IntegerToString(k) cat "|" cat IntegerToString(d) cat "|"
              cat intseq_to_str(B : sep:=",");
    // Is the singularity cached?
    bool,Q:=StoreIsDefined(sing_cache,key);
    if not bool then
        // No - create a new singularity with the given data
        Q:=New(FmtSng);
        Q`index:=r;
        Q`weights:=w;
        Q`polarisation:=k;
        Q`dimension:=d;
        if #B gt 0 then Q`eqn_weights:=B; end if;
        // Assign an id
        bool,id:=StoreIsDefined(sing_cache,"id");
        id:=bool select id + 1 else 1;
        Q`id:=id;
        // Update the cache
        StoreSet(sing_cache,"id",id);
        StoreSet(sing_cache,key,Q);
    end if;
    // Return the singularity
    return Q;
end function;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// A comparison function for two singularities Q1 and Q2. Returns -'ve if
// Q1 < Q2, 0 if Q1 = Q2, and +'ve if Q1 > Q2.
function cmp_singularity(Q1,Q2)
    // Are the ids the same?
    if Q1`id eq Q2`id then return 0; end if;
    // Determine an order
    d:=Dimension(Q1) - Dimension(Q2);
    if d ne 0 then return d; end if;
    d:=Index(Q1) - Index(Q2);
    if d ne 0 then return d; end if;
    d:=GlobalPolarisation(Q1) - GlobalPolarisation(Q2);
    if d ne 0 then return 0; end if;
    W1:=Weights(Q1);
    W2:=Weights(Q2);
    if W1 ne W2 then
        return W1 gt W2 select 1 else -1;
    end if;
    E1:=EquationWeights(Q1);
    E2:=EquationWeights(Q2);
    if E1 ne E2 then
        return E1 gt E2 select 1 else -1;
    end if;
    // If we're here then we can't tell the difference between them
    return 0;
end function;

// Check that the quotient singularity 1/r(a_1,...,a_n) -- we write
// A = [a_1,...,a_n] -- is smooth in codimension 1; i.e. that at least two
// a_i are coprime to r.
function smooth_in_codim1(r,A)
    found_one:=false;
    for a in A do
        if IsCoprime(r,a) then
            if found_one then
                return true;
            else
                found_one:=true;
            end if;
        end if;
    end for;
    return false;
end function;

// Compute the discrepancies of a resolution of a toric quotient singularity
// 1/r(a_1,...,a_n) - we write A = [a_1,...,.a_n].
function discrepancies(Q)
    r:=Index(Q);
    A:=Weights(Q);
    dim:=#A;
    C:=PositiveQuadrant(dim);
    L:=Ambient(C);
    v:=L ! [a / r : a in A];
    _,emb:=AddVectorToLattice(v);
    CC:=Image(emb,C);
    u:=Dual(L) ! [1 : i in [1..Dimension(L)]];  // toric -K_X for C^n
    uu:=u @@ Dual(emb);
    corners:=[emb(w) : w in Rays(C)];
    return [Rationals() | uu * w - 1 : w in ZGenerators(CC :
                                     algorithm:="Hemmecke") | w notin corners];
end function;

// Returns the numerator of the Q_orb contribution for the case 1/r(-1,a,-a).
function ice_cream_gt3(r,a)
    // Find m; repick a if necessary
    m:=Integers() ! ((Integers(r) ! (-a))^(-1));
    if m gt r/2 then
        a:=r - a;
        m:=Integers() ! ((Integers(r) ! (-a))^(-1));
    end if;
    // Build the numerator coefficients
    coeffs:=[Integers()|];
    for i in [0..r - 2] do
        ia:=(-i*m) mod r;
        Append(~coeffs, -Minimum([m,AbsoluteValue(m - ia)]));
    end for;
    // Finally, construct the numerator as a polynomial in \QQ[t]
    R:=PolynomialRing(Rationals());
    return &+[R | coeffs[i + 1] * R.1^(i + 3) : i in [0..r - 2]];
end function;

// Returns the numerator of the Q_orb contribution.
// See Buckley, Reid, Zhou 2012. Code from Reid's webpage:
//      http://www.warwick.ac.uk/~masda/Ice/My_magmarc
function qorb_numerator(r,L,k,n)
    // Sanity check
    assert n eq #L;
    ChangeUniverse(~L,Integers());
    assert (k + &+L) mod r eq 0;
    // Create \QQ[t]
    R:=PolynomialRing(Rationals());
    t:=R.1;
    // Do the stuff
    Pi:=&*[R | 1 - t^i : i in L];
    A:=(1 - t^r) div (1 - t);
    B:=Pi div (1 - t)^n;
    H:=GCD(A,B);
    M:=&*[GCD(A, 1 - t^i) : i in L];
    shift:=Floor(Degree(M * H) / 2);
    l:=Floor((k + n + 1) / 2 + shift + 1);
    // If l < 0 we need a kludge to avoid programming Laurent polynomials
    de:=Maximum(0,Ceiling(-l / r));
    m:=l + de * r;
    G,_,be:=XGCD(A div H, t^m * B div M);
    return t^m * be / (M * (1 - t)^n * (1 - t^r) * t^(de * r));
end function;

// True iff the properties given by the sequence S are valid. Otherwise returns
// false followed by an error message.
function check_properties(S)
    for t in S do
        if t ne "quotient" and t ne "terminal" and t ne "logterminal" and
           t ne "canonical" and t ne "logcanonical" and t ne "isolated" and
           t ne "nonsingular" and t ne "smooth" and t ne "gorenstein" then
            return false,"Unknown property \"" cat t cat "\".";
        end if;
    end for;
    return true,_;
end function;

// True iff the singularity Q has all properties given by the sequence S
// (e.g. {"isolated", "terminal"}).
function singularity_is_of_type(Q,S)
    bool:=true;
    // The order that we check this list is important!
    if "quotient" in S then
        bool and:= IsQuotientSingularity(Q);
    end if;
    if "isolated" in S then
        bool and:= IsIsolatedSingularity(Q);
    end if;
    if "terminal" in S then
        bool and:= IsTerminalSingularity(Q);
    end if;
    if "canonical" in S then
        bool and:= IsCanonicalSingularity(Q);
    end if;
    if "logterminal" in S then
        bool and:= IsLogTerminalSingularity(Q);
    end if;
    if "logcanonical" in S then
        bool and:= IsLogTerminalSingularity(Q);
    end if;
    if "gorenstein" in S then
        bool and:= IsGorenstein(Q);
    end if;
    return bool;
end function;

/////////////////////////////////////////////////////////////////////////
// Creation intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Singularity(r::RngIntElt,A::SeqEnum[RngIntElt],k::RngIntElt) -> FmtSng
{Create a quotient singularity of index r and weights A with global polarisation k}
    // Sanity check
    require r ge 1: "Argument 1 must be a positive integer";
    if IsNull(A) then
        A:=[Integers()|];
    else
        bool,A:=CanChangeUniverse(A,Integers());
        require bool: "Argument 2 must be a sequence of integers";
    end if;
    require (k + &+A) mod r eq 0: "Canonical weight not compatible";
    // Return the singularity
    return get_sing_from_cache(r,A,k,#A,[Integers()|]);
end intrinsic;

intrinsic Singularity(r::RngIntElt,A::SeqEnum[RngIntElt]) -> FmtSng
{Create a quotient singularity of index r and weights A}
    // Sanity check
    require r ge 1: "Argument 1 must be a positive integer";
    if IsNull(A) then
        A:=[Integers()|];
    else
        bool,A:=CanChangeUniverse(A,Integers());
        require bool: "Argument 2 must be a sequence of integers";
    end if;
    // Return the singularity
    k:=(-&+A) mod r;
    return get_sing_from_cache(r,A,k,#A,[Integers()|]);
end intrinsic;

intrinsic Singularity(d::RngIntElt,r::RngIntElt,A::SeqEnum[RngIntElt],
                      B::SeqEnum[RngIntElt],k::RngIntElt) -> FmtSng
{Create a (hyperquotient) singularity of dimension d, index r and weights A and local equation weights B and global polarisation k}
    // Note: We do not check that A,B,k are compatible! We can't unless we
    // know, for example, that this is a local complete intersection
    // hyperquotient.
    // Sanity check
    require d ge 1: "Argument 1 must be a positive integer";
    require d eq #A - #B:
        "Arguments do not determine a quotient of a complete intersection";
    require r ge 1: "Argument 2 must be a positive integer";
    if IsNull(A) then
        A:=[Integers()|];
    else
        bool,A:=CanChangeUniverse(A,Integers());
        require bool: "Argument 3 must be a sequence of integers";
    end if;
    if IsNull(B) then
        B:=[Integers()|];
    else
        bool,B:=CanChangeUniverse(B,Integers());
        require bool: "Argument 4 must be a sequence of integers";
    end if;
    require (k mod r) eq ((&+B - &+A) mod r):
        "Local canonical class is not correct";
    // Return the singularity
    return get_sing_from_cache(r,A,k,d,B);
end intrinsic;

intrinsic AreWeightsCompatible(r::RngIntElt,A::SeqEnum[RngIntElt],k::RngIntElt)
    -> BoolElt,FmtSng
{True iff 1/r(A) is a quotient singularity polarised in degree k. In that case, also gives the corresponding quotient singularity of index r and weights A.}
    // Sanity check
    require r gt 0: "Argument 1 must be a positive integer";
    if IsNull(A) then A:=[Integers()|]; end if;
    // Return the result
	if (k + &+A) mod r eq 0 then
		return true,get_sing_from_cache(r,A,k,#A,[Integers()|]);
	else
		return false,_;
	end if;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// Display intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Print(Q::FmtSng,l::MonStgElt)
{Print the singularity Q at print level l}
    if l ne "Minimal" and l ne "Default" and l ne "Maximal" and l ne "Magma" and
       l ne "Latex" then l:="Default"; end if;
    printf SingularityToString(Q : level:=l);
end intrinsic;

intrinsic SingularityToString(Q::FmtSng : level:="Default") -> MonStgElt
{A string representation of the singularity Q}
    // Sanity check
    require Type(level) eq MonStgElt and (level eq "Minimal" or
        level eq "Default" or level eq "Maximal" or level eq "Magma" or
        level eq "Latex"):
        "Parameter 'level' should be one of \"Minimal\", \"Default\", \"Maximal\", \"Magma\", or \"Latex\"";
    // Is this the minimal print level?
	if level eq "Minimal" then
        str:="1/" cat IntegerToString(Index(Q)) cat "(" cat
                                              intseq_to_str(Weights(Q)) cat ")";
    // Is this the LaTeX print level?
    elif level eq "Latex" then
        B:=EquationWeights(Q);
        if #B eq 0 then
            str:="\\frac{1}{" cat IntegerToString(Index(Q)) cat "}("
                                          cat intseq_to_str(Weights(Q)) cat ")";
        else
            str:="\\frac{1}{" cat IntegerToString(Index(Q)) cat "}("
                                          cat intseq_to_str(Weights(Q)) cat ";"
                                          cat intseq_to_str(B) cat ")";
        end if;
    // Is this the Magma print level?
    elif level eq "Magma" then
        r:=Index(Q);
		A:=Weights(Q);
		k:=GlobalPolarisation(Q);
		B:=EquationWeights(Q);
        if #B then
            str:="Singularity(" cat IntegerToString(Dimension(Q)) cat ", " cat
                           IntegerToString(r) cat ", [" cat intseq_to_str(A) cat
                           "], [" cat intseq_to_str(B) cat "], " cat
                           IntegerToString(k) cat ")";
        elif k eq (-&+A) mod r then
            str:="Singularity(" cat IntegerToString(r) cat ", [" cat
                                                      intseq_to_str(A) cat "])";
        else
            str:="Singularity(" cat IntegerToString(r) cat ", [" cat
                      intseq_to_str(A) cat "], " cat IntegerToString(k) cat ")";
        end if;
	else
        // Create the default print data
        W:=Weights(Q);
        if #W eq 0 then
            str:="Curve with 1/" cat IntegerToString(Index(Q));
        else
            if IsQuotientSingularity(Q) then
                str:="1/" cat IntegerToString(Index(Q)) cat "(" cat
                                              intseq_to_str(Weights(Q)) cat ")";
            else
                str:="1/" cat IntegerToString(Index(Q)) cat "(" cat
                                      intseq_to_str(Weights(Q)) cat "; " cat
                                      intseq_to_str(EquationWeights(Q)) cat ")";
            end if;
        end if;
        // Do we need to append the maximal data?
        if level eq "Maximal" then
            if assigned Q`eqn_weights and #EquationWeights(Q) gt 0 then
                if #EquationWeights(Q) eq 1 then
                    str cat:= " with equation weight ";
                else
                    str cat:= " with equation weights ";
                end if;
                str cat:= intseq_to_str(EquationWeights(Q));
            end if;
        end if;
    end if;
    // Return the string
    return str;
end intrinsic;

//////////////////////////////////////////////////////////////////////
// Comparison and hash intrinsics
//////////////////////////////////////////////////////////////////////

intrinsic Hash(Q::FmtSng) -> RngIntElt
{The hash value of the singularity Q}
    if not assigned Q`hash then
        hash:=(Hash(Dimension(Q)) + 17 * Hash(Index(Q))) mod 2^32;
        hash:=(hash + 17^2 * Hash(GlobalPolarisation(Q))) mod 2^32;
        hash:=(hash + 17^3 * Hash(Weights(Q))) mod 2^32;
        Q`hash:=hash;
    end if;
    return Q`hash;
end intrinsic;

intrinsic 'eq'(Q1::FmtSng,Q2::FmtSng) -> BoolElt
{True if the two singularities Q1 and Q2 are equal, false otherwise}
    return cmp_singularity(Q1,Q2) eq 0;
end intrinsic;

intrinsic 'lt'(Q1::FmtSng,Q2::FmtSng) -> BoolElt
{True if the singularity Q1 < Q2, false otherwise}
    return cmp_singularity(Q1,Q2) lt 0;
end intrinsic;

intrinsic 'le'(Q1::FmtSng,Q2::FmtSng) -> BoolElt
{True if the singularity Q1 <= Q2, false otherwise}
    return cmp_singularity(Q1,Q2) le 0;
end intrinsic;

intrinsic Sort(Q::SeqEnum[FmtSng]) -> SeqEnum[FmtSng]
{Sorts the sequence Q of singularities}
    // Get the untyped case out of the way
    if IsNull(Q) then return Q; end if;
    // Sort the sequence
    Sort(~Q,cmp_singularity);
    return Q;
end intrinsic;

intrinsic Sort(~Q::SeqEnum[FmtSng])
{Sorts in place the sequence Q of singularities}
    Sort(~Q,cmp_singularity);
end intrinsic;

intrinsic Sort(S::SetIndx[FmtSng]) -> SeqEnum[FmtSng]
{Sorts the indexed set S of singularities}
    // Get the untyped case out of the way
    if IsNull(S) then return S; end if;
    // Sort the indexed set
    Sort(~S,cmp_singularity);
    return S;
end intrinsic;

intrinsic Sort(~S::SetIndx[FmtSng])
{Sort the indexed set S of singularities in place}
    Sort(~S,cmp_singularity);
end intrinsic;

intrinsic DistinctSingularities(Q::SeqEnum[FmtSng])
    -> SeqEnum[FmtSng],SeqEnum[RngIntElt]
{The sequence of distinct singularities in the sequence Q, along with the indices of a corresponding representative in Q}
    // Get the untyped case out of the way
    if IsNull(Q) then return Q,[Integers()|]; end if;
    // Extract the distinct singularities
    S:=[Universe(Q)|];
    idxs:=[Integers()|];
    for i in [1..#Q] do
        if not &or[cmp_singularity(Q[i],Q[j]) eq 0 : j in [i + 1..#Q]] then
            Append(~S,Q[i]);
            Append(~idxs,i);
        end if;
    end for;
    // Sort and return the sequence
    Sort(~S,cmp_singularity,~permut);
    idxs:=PermuteSequence(idxs,permut);
    return S,idxs;
end intrinsic;

//////////////////////////////////////////////////////////////////////
// Access intrinsics
//////////////////////////////////////////////////////////////////////

intrinsic Dimension(Q::FmtSng) -> RngIntElt
{The dimension of the singularity Q}
    return Q`dimension;
end intrinsic;

intrinsic Index(Q::FmtSng) -> RngIntElt
{The index r of the singularity Q}
    return Q`index;
end intrinsic;

intrinsic Weights(Q::FmtSng) -> SeqEnum[RngIntElt]
{The polarising weights (taken modulo the index) of the singularity Q}
    return Q`weights;
end intrinsic;

intrinsic EquationWeights(Q::FmtSng) -> SeqEnum[RngIntElt]
{The weights of the local defining equations (taken modulo the index) of the singularity Q}
    return assigned Q`eqn_weights select Q`eqn_weights else [Integers()|];
end intrinsic;

intrinsic GlobalPolarisation(Q::FmtSng) -> RngIntElt
{The global (integer) polarisation k of the singularity Q. (The given
polarisation is consistent with Q lying on a variety X in weighted
projective space so that degree k sections of X are canonical divisors.)}
    return Q`polarisation;
end intrinsic;

intrinsic Polarisation(Q::FmtSng) -> RngIntElt
{The local (integral residue mod the index) polarisation of the singularity Q}
    return GlobalPolarisation(Q) mod Index(Q);
end intrinsic;

intrinsic OrbifoldHilbertNumerator(Q::FmtSng) -> FldFunRatUElt
{The numerator of the local orbifold Hilbert Series contribution of the isolated quotient singularity Q}
    // Do we need to compute the numerator?
    if not assigned Q`porb_num then
        // Note: We can currently only work with isolated quotient singularities
        require IsQuotientSingularity(Q):
            "Only defined for quotient singularities";
        A:=Weights(Q);
        r:=Index(Q);
        k:=GlobalPolarisation(Q);
        d:=Dimension(Q);
        if d eq 3 and k eq 1 and GCD(A[1],r) eq 1 and A[3] eq r - 1 then
            // This is the case 1/r(a,r-a,r-1) for terminal canonical 3-folds.
            Q`porb_num:=ice_cream_gt3(r,A[1]);
	else
            // qorb_numerator actually computes the whole series GDB Jan2013 THINK
	    P := qorb_numerator(r,A,k,d);
            Q`porb:=P;
	    t := Parent(P).1;
            Q`porb_num:= P * (1-t)^#A*(1-t^r);
	end if;
    end if;
    // Return the numerator
    return Q`porb_num;
end intrinsic;

intrinsic OrbifoldHilbertSeries(Q::FmtSng) -> FldFunRatUElt
{The local orbifold Hilbert Series contribution of the quotient singularity Q}
    // Do we need to compute the Q_orb contribution?
    if not assigned Q`porb then
        // Note: We can currently only work with isolated quotient singularities
        require IsQuotientSingularity(Q):
            "Only defined for quotient singularities";
        pnum:=OrbifoldHilbertNumerator(Q);
        R:=Parent(pnum);
        Q`porb:=pnum / ((1 - R.1)^Dimension(Q) * (1 - R.1^Index(Q)));
    end if;
    // Return the contribution
    return Q`porb;
end intrinsic;

intrinsic OrbifoldHilbertCoefficients(Q::FmtSng,n::RngIntElt)
    -> SeqEnum[RngIntElt]
{The first n coefficients of the local orbifold Hilbert Series contribution of the quotient singularity Q (i.e. coefficients of 1,t,...,t^(n-1))}
    // Sanity check
    require n gt 0: "Argument 2 must be a positive integer";
    // We keep a cache of the first 20 coefficients -- can we use it?
    if n le 20 then
        if not assigned Q`first20coeffs then
            h:=OrbifoldHilbertSeries(Q);
            Q`first20coeffs:=FirstNCoefficients(h,20);
        end if;
        coeffs:=Q`first20coeffs[1..n];
    else
        h:=OrbifoldHilbertSeries(Q);
        coeffs:=FirstNCoefficients(h,n);
        if not assigned Q`first20coeffs then
            Q`first20coeffs:=coeffs[1..20];
        end if;
    end if;
    // Return the coefficients
    return coeffs;
end intrinsic;

intrinsic OrbifoldHilbertCoefficientsInRange(Q::FmtSng,n1::RngIntElt,
    n2::RngIntElt) -> SeqEnum[RngIntElt]
{The coefficients of t^n1,..,t^n2 of the local orbifold Hilbert Series contribution of the quotient singularity Q}
    // Sanity check
    require n1 le n2: "Argument 3 must be greater than or equal to argument 2";
    // Return the coefficients
    return CoefficientsInRange(OrbifoldHilbertSeries(Q),n1,n2);
end intrinsic;

intrinsic Plurigenus(Q::FmtSng,n::RngIntElt) -> RngIntElt
{The coefficient of t^n in the local orbifold Hilbert series of the quotient singularity Q}
    // Sanity check
    require n ge 0: "Argument 2 must be non-negative";
    // We keep a cache of the first 20 coefficients -- can we use it?
    if n lt 20 then
        coeffs:=OrbifoldHilbertCoefficients(Q,n + 1);
        if #coeffs ge n+1 then return coeffs[n+1]; end if;
    end if;
    // We need to compute the n-th coefficient in the power series ring
    S:=PowerSeriesRing(Rationals(),n + 1);
    return Coefficient(S ! OrbifoldHilbertSeries(Q),n);
end intrinsic;

//////////////////////////////////////////////////////////////////////
// Singularity analysis intrinsics
//////////////////////////////////////////////////////////////////////

intrinsic IsQuotientSingularity(Q::FmtSng) -> BoolElt
{True if and only if the singularity Q is a quotient singularity (that is, there are no equations as part of its defining data)}
    return #EquationWeights(Q) eq 0;
end intrinsic;

intrinsic IsIsolatedSingularity(Q::FmtSng) -> BoolElt
{True iff the (quotient) singularity Q is isolated.}
    // Note: We can currently only work with quotient singularities
    require IsQuotientSingularity(Q):
        "Only quotient singularities are currently supported";
    return &and[GCD(Index(Q),w) eq 1 : w in Weights(Q)];
end intrinsic;

intrinsic IsLogCanonicalSingularity(Q::FmtSng) -> BoolElt
{True iff the (quotient) singularity Q is log canonical.}
    if not assigned Q`is_logcanonical then
        require IsQuotientSingularity(Q) and
            smooth_in_codim1(Index(Q),Weights(Q)):
            "Only quotient singularities smooth in codim 1 are supported";
        if Dimension(Q) eq 2 then	// always log terminal
            Q`is_logcanonical:=true;
            Q`is_logterminal:=true;
        else
            D:=discrepancies(Q);
            Q`is_logcanonical:=&and[a ge -1 : a in D];
            Q`is_logterminal:=&and[ a gt -1 : a in D];
            Q`is_canonical:=&and[a ge 0 : a in D];
            Q`is_terminal:=&and[a gt 0 : a in D];
        end if;
    end if;
    return Q`is_logcanonical;
end intrinsic;

intrinsic IsGorenstein(Q::FmtSng) -> BoolElt
{True iff the (quotient) singularity Q is Gorenstein.}
    // Note: We can currently only work with quotient singularities
    require IsQuotientSingularity(Q):
        "Only quotient singularities are currently supported";
    bool := IsDivisibleBy(&+[ w : w in Weights(Q)],Index(Q));
    return bool;
end intrinsic;

intrinsic IsLogTerminalSingularity(Q::FmtSng) -> BoolElt
{True iff the (quotient) singularity Q is log terminal.}
    if not assigned Q`is_logterminal then
        require IsQuotientSingularity(Q) and
            smooth_in_codim1(Index(Q),Weights(Q)):
            "Only quotient singularities smooth in codim 1 are supported";
        if Dimension(Q) eq 2 then	// always log terminal
            Q`is_logcanonical:=true;
            Q`is_logterminal:=true;
        else
            D:=discrepancies(Q);
            Q`is_logcanonical:=&and[a ge -1 : a in D];
            Q`is_logterminal:=&and[a gt -1 : a in D];
            Q`is_canonical:=&and[a ge 0 : a in D];
            Q`is_terminal:=&and[a gt 0 : a in D];
        end if;
    end if;
    return Q`is_logterminal;
end intrinsic;

intrinsic IsTerminalSingularity(Q::FmtSng) -> BoolElt
{True iff the (quotient) singularity Q is terminal.}
    if not assigned Q`is_terminal then
        // Note: We can currently only work with quotient singularities
        require IsQuotientSingularity(Q) and
            smooth_in_codim1(Index(Q),Weights(Q)):
            "Only quotient singularities smooth in codim 1 are supported";
        // Use the Reid, Shepherd-Barron, Tai criterion
        r:=Index(Q);
        Q`is_terminal:=&and[&+[Integers() | (i*a) mod r : a in Weights(Q)] gt r
                                                               : i in [1..r-1]];
        if Q`is_terminal then
            Q`is_canonical:=true;
            Q`is_logterminal:=true;
            Q`is_logcanonical:=true;
        end if;
    end if;
    return Q`is_terminal;
end intrinsic;

intrinsic IsCanonicalSingularity(Q::FmtSng) -> BoolElt
{True iff the (quotient) singularity Q is canonical.}
    if not assigned Q`is_canonical then
        // Note: We can currently only work with quotient singularities
        require IsQuotientSingularity(Q) and
                smooth_in_codim1(Index(Q),Weights(Q)):
        "Only quotient singularities smooth in codim 1 are supported";
        // Use the Reid, Shepherd-Barron, Tai criterion
	    r:=Index(Q);
        Q`is_canonical:=&and[&+[Integers() | (i*a) mod r : a in Weights(Q)] ge r
                                                               : i in [1..r-1]];
        if not Q`is_canonical then Q`is_terminal:=false; end if;
        if Q`is_canonical then
            Q`is_logterminal:=true;
            Q`is_logcanonical:=true;
        end if;
    end if;
    return Q`is_canonical;
end intrinsic;

intrinsic IsOfType(Q::FmtSng,S::MonStgElt) -> BoolElt
{True iff the singularity Q has property S (e.g. "isolated, terminal").}
    S:=split_to_lc(S);
    bool,msg:=check_properties(S);
    require bool: msg;
    return singularity_is_of_type(S);
end intrinsic;
