freeze;

/////////////////////////////////////////////////////////////////////////
// strings.m
/////////////////////////////////////////////////////////////////////////
// A collection of useful utility functions for working with strings.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
// num_to_latex
/////////////////////////////////////////////////////////////////////////
// Input: An integer or rational n.
// Output: A LaTeX string for n.
/////////////////////////////////////////////////////////////////////////

function num_to_latex(n)
    return IsIntegral(n) select IntegerToString(Integers() ! n)
                    else  Sprintf("\\frac{%o}{%o}",Numerator(n),Denominator(n));
end function;

/////////////////////////////////////////////////////////////////////////
// split_to_lc
/////////////////////////////////////////////////////////////////////////
// Input: A string S.
// Output: A set SS given by splitting S at ',' and ' ' and
// converting the resulting strings to lower case.
/////////////////////////////////////////////////////////////////////////
/*
S:="Isolated Quotient, Terminal";
split_to_lc(S);  // { "quotient", "terminal", "isolated" }
*/

function split_to_lc(S)
    return {StringToLower(s) : s in Split(S," ,")};
end function;

/////////////////////////////////////////////////////////////////////////
// intseq_to_str
/////////////////////////////////////////////////////////////////////////
// Input: A sequence of integers.
// Ouput: A string representation without brackets.
/////////////////////////////////////////////////////////////////////////
/*
S:=[ 1, 3, 5 ];
intseq_to_str(S);   // "1, 3, 5"
*/

// Given a sequence of integers, returns the string representation (without
// brackets).
function intseq_to_str(S : sep:=", ")
    if #S eq 0 then return ""; end if;
    str:="";
    for c in S do
        str cat:= IntegerToString(c) cat sep;
    end for;
    return str[1..#str - #sep];
end function;

/////////////////////////////////////////////////////////////////////////
// check_bracket_type, check_matrix_brackets_type, check_seperator_type,
// get_brackets, get_matrix_brackets, get_seperator
/////////////////////////////////////////////////////////////////////////
// Internal use function to get the bracket information and seperator
// information.
/////////////////////////////////////////////////////////////////////////

function check_bracket_type(brackets)
    if Type(brackets) ne MonStgElt then
        return false,"Bracket style must be a string.";
    end if;
    if brackets ne "[" and brackets ne "(" and brackets ne "{" and
       brackets ne "|" and brackets ne "<" and brackets ne "" and
       brackets ne " " then
        return false,"Bracket style is unknown. Should be one of \"[\", \"(\", \"{\", \"|\", or \"\" for no brackets.";
    end if;
    return true,_;
end function;

function check_matrix_bracket_type(brackets)
    if Type(brackets) ne MonStgElt then
        return false,"Bracket style must be a string.";
    end if;
    if brackets ne "[" and brackets ne "(" and brackets ne "{" and
       brackets ne "|" and brackets ne "" and brackets ne " " then
        return false,"Bracket style is unknown. Should be one of \"[\", \"(\", \"{\", \"|\", or \"\" for no brackets.";
    end if;
    return true,_;
end function;

function check_seperator_type(sep)
    if Type(sep) ne MonStgElt then
        return false,"Seperator style must be a string.";
    end if;
    if sep ne "," and sep ne ":" and sep ne " " and sep ne "" then
        return false,"Seperator style is unknown. Should be one of \",\" and \":\", or \"\" for no seperator.";
    end if;
    return true,_;
end function;

function get_brackets(brackets,latex)
    // Sanity check
    bool,msg:=check_bracket_type(brackets);
    error if not bool, msg;
    // Compute the left and right brackets
    idx := Index(["[","(","{","|","<",""," "],brackets);
    assert idx ne 0;
    if idx lt 6 then
        if latex then
            left := ["\\left[","\\left(","\\left\\{","\\left|","\\left<"][idx];
            right := ["\\right]","\\right)","\\right\\}","\\right|","\\right>"][idx];
        else
            left := ["[","(","{", "|","<"][idx];
            right := ["]",")","}","|",">"][idx];
        end if;
        brackets := true;
    else
        left := "";
        right := "";
        brackets := false;
    end if;
    return brackets,left,right;
end function;

function get_matrix_brackets(brackets)
    // Sanity check
    bool,msg:=check_matrix_bracket_type(brackets);
    error if not bool, msg;
    // Compute the left and right brackets
    idx := Index(["[","(","{","|",""," "],brackets);
    error if idx eq 0, "Bracket style is unknown";
    return ["bmatrix","pmatrix","Bmatrix","vmatrix","matrix","matrix"][idx];
end function;

function get_seperator(sep)
    // Sanity check
    bool,msg:=check_seperator_type(sep);
    error if not bool, msg;
    // Compute the seperator
    idx := Index([",",":"," ",""],sep);
    error if idx eq 0, "Seperator style is unknown";
    sep := [", "," : "," "," "][idx];
    return sep;
end function;

/////////////////////////////////////////////////////////////////////////
// weights_to_string
/////////////////////////////////////////////////////////////////////////
// Input: A sequence Q of integer weights. Set 'brackets' to one of
// "[" (default), "(", "{", "|", or "" to suppress brackets.
// Output: A string representation of Q, with the weights sorted in
// increasing order, with multiplicities recorded as superscripts.
/////////////////////////////////////////////////////////////////////////

function weights_to_string(Q : brackets:="[", latex:=false )
    // Sanity check
    bool,msg:=check_bracket_type(brackets);
    error if not bool, msg;
    error if Type(latex) ne BoolElt, "Parameter 'latex' must be a boolean";
    // Get the bracket and seperator
    brackets,left,right:=get_brackets(brackets,latex);
    sep:=get_seperator(",");
    // Is there anything to do?
    if #Q eq 0 then
        return brackets select left cat " " cat right else "";
    end if;
    // Construct the string
    s:="";
    for w in Sort(SetToSequence(SequenceToSet(Q))) do
        n := #[Integers() | 1 : i in Q | i eq w];
        if n ge 10 and latex then
            s cat:= IntegerToString(w) cat "^{" cat IntegerToString(n) cat
                                                                    "}" cat sep;
        elif n eq 1 then
            s cat:= IntegerToString(w) cat sep;
        else
            s cat:= IntegerToString(w) cat "^" cat IntegerToString(n) cat sep;
        end if;
    end for;
    return brackets select left cat s[1..#s-#sep] cat right
                      else s[1..#s-#sep];
end function;

/////////////////////////////////////////////////////////////////////////
// strict_upper_triangular_matrix_to_string
/////////////////////////////////////////////////////////////////////////
// Input: A square matrix 'A'. Set 'brackets' to one of "[",
// "(" (default), "{", "|", or "" to suppress brackets. 'prefix' will be
// prefixed to each row except for the first row.
// Output: A string representation of the strict upper triangular
// entries of 'A'.
/////////////////////////////////////////////////////////////////////////

function strict_upper_matrix_to_latex(A,brackets)
    // Get bracket type
    type:=get_matrix_brackets(brackets);
    // Get the number of columns/rows
    num_cols:=NumberOfColumns(A);
    // Create the string
    str:="\\begin{" cat type cat "} ";
    for i in [1..num_cols - 1] do
        if i ne 1 then
            str cat:= &cat["&" : j in [1..i - 1]];
        end if;
        for j in [i + 1..num_cols] do
            if j ne num_cols then
                str cat:= num_to_latex(A[i][j]) cat "&";
            else
                if i ne num_cols - 1 then
                    str cat:= num_to_latex(A[i][j]) cat "\\\\ ";
                else
                    str cat:= num_to_latex(A[i][j]) cat " ";
                end if;
            end if;
        end for;
    end for;
    return str cat "\\end{" cat type cat "}";
end function;

function strict_upper_matrix_to_string(A,brackets,prefix)
    // Get the number of columns/rows
    num_cols:=NumberOfColumns(A);
    // Get the bracket and seperator
    brackets,left,right:=get_brackets(brackets,false);
    sep:=get_seperator("");
    // Convert the entries to strings and calculate the maximum string width
    // for each column
    S:=[];
    widths:=ZeroSequence(Integers(),num_cols - 1);
    for i in [1..num_cols - 1] do
        for j in [i + 1..num_cols] do
            s:=Sprintf("%o",A[i][j]);
            Append(~S,s);
            if #s gt widths[j - 1] then widths[j - 1]:=#s; end if;
        end for;
    end for;
    // Create the string
    str:="";
    idx:=1;
    for i in [1..num_cols - 1] do
        if i ne 1 then
            if brackets then
                str cat:= prefix cat left;
            else
                str cat:= prefix;
            end if;
        elif brackets then
            str cat:= left;
        end if;
        for j in [1..i - 1] do
            str cat:= &cat[" " : k in [1..widths[j]]] cat " ";
        end for;
        for j in [i..num_cols - 1] do
            diffs:=widths[j] - #S[idx];
            if diffs gt 0 then
                str cat:= &cat[" " : k in [1..diffs]];
            end if;
            if j eq num_cols - 1 then
                if i eq num_cols - 1 then
                    if brackets then
                        str cat:= S[idx] cat right;
                    else
                        str cat:= S[idx];
                    end if;
                else
                    if brackets then
                        str cat:= S[idx] cat right cat "\n";
                    else
                        str cat:= S[idx] cat "\n";
                    end if;
                end if;
            else
                str cat:= S[idx] cat sep;
            end if;
            idx +:= 1;
        end for;
    end for;
    return str;
end function;

function strict_upper_triangular_matrix_to_string( A : brackets:="(",
    latex:=false, prefix:="" )
    // Sanity check
    error if NumberOfRows(A) ne NumberOfColumns(A), "The matrix must be square";
    error if Type(latex) ne BoolElt, "Parameter 'latex' must be a boolean";
    bool,msg:=check_matrix_bracket_type(brackets);
    error if not bool, msg;
    error if Type(prefix) ne MonStgElt, "Parameter 'prefix' must be a string";
    // For ease we call a helper functions
    return latex select strict_upper_matrix_to_latex(A,brackets)
                   else strict_upper_matrix_to_string(A,brackets,prefix);
end function;

/////////////////////////////////////////////////////////////////////////
// matrix_to_string
/////////////////////////////////////////////////////////////////////////
// Input: A matrix 'A'. Set 'brackets' to one of "[", "(" (default),
// "{", "|", or "" to suppress brackets. 'prefix' will be prefixed to
// each row except for the first row. If 'latex' is true then 'prefix'
// will be ignored.
// Output: A string representation of the matrix 'A'.
/////////////////////////////////////////////////////////////////////////

function matrix_to_latex(A,brackets)
    // Get bracket type
    type:=get_matrix_brackets(brackets);
    // Record the matrix size
    num_cols:=NumberOfColumns(A);
    num_rows:=NumberOfRows(A);
    // Create the string
    str:="\\begin{" cat type cat "} ";
    for i in [1..num_rows] do
        for j in [1..num_cols] do
            if j ne num_cols then
                str cat:= num_to_latex(A[i][j]) cat "&";
            else
                if i ne num_rows then
                    str cat:= num_to_latex(A[i][j]) cat "\\\\ ";
                else
                    str cat:= num_to_latex(A[i][j]) cat " ";
                end if;
            end if;
        end for;
    end for;
    return str cat "\\end{" cat type cat "}";
end function;

function matrix_to_string(A,brackets,prefix)
    // Get the bracket and seperator
    brackets,left,right:=get_brackets(brackets,false);
    sep:=get_seperator("");
    // Convert the integers to strings a row at a time and calculate the
    // maximum column string width as we go along
    num_cols:=NumberOfColumns(A);
    num_rows:=NumberOfRows(A);
    S:=[];
    widths:=ZeroSequence(Integers(),num_cols);
    for i in [1..num_rows] do
        for j in [1..num_cols] do
            s:=Sprintf("%o",A[i][j]);
            Append(~S,s);
            if #s gt widths[j] then
                widths[j]:=#s;
            end if;
        end for;
    end for;
    // Create the string
    str:="";
    idx:=1;
    for i in [1..num_rows] do
        if i ne 1 then
            if brackets then
                str cat:= prefix cat left;
            else
                str cat:= prefix;
            end if;
        elif brackets then
            str cat:= left;
        end if;
        for j in [1..num_cols] do
            diffs:=widths[j] - #S[idx];
            if diffs gt 0 then
                str cat:= &cat[" " : k in [1..diffs]];
            end if;
            if j eq num_cols then
                if i eq num_rows then
                    if brackets then
                        str cat:= S[idx] cat right;
                    else
                        str cat:= S[idx];
                    end if;
                else
                    if brackets then
                        str cat:= S[idx] cat right cat "\n";
                    else
                        str cat:= S[idx] cat "\n";
                    end if;
                end if;
            else
                str cat:= S[idx] cat sep;
            end if;
            idx +:= 1;
        end for;
    end for;
    return str;
end function;

function matrix_to_string(A : brackets:="(", latex:=false, prefix:="")
    // Sanity check
    error if Type(latex) ne BoolElt, "Parameter 'latex' must be a boolean";
    bool,msg:=check_matrix_bracket_type(brackets);
    error if not bool, msg;
    error if Type(prefix) ne MonStgElt, "Parameter 'prefix' must be a string";
    // For ease we call a helper function
    return latex select matrix_to_latex(A,brackets)
                   else matrix_to_string(A,brackets,prefix);
end function;
