//freeze;

/////////////////////////////////////////////////////////////////////////
// ci.m
/////////////////////////////////////////////////////////////////////////
// Complete intersection format.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////

import "utilities.m": search_format;
import "../Engine/strings.m": intseq_to_str;
import "../Engine/utilities.m": knapsack_with_repeats;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// l is the print level; it can be used to distinguish levels:
// "Minimal", "Default" or "Maximal".
function extra_to_string(F,l)
    return "";  // Nothing useful to add
end function;

// Additional data about the format F as a LaTeX formatted string.
function extra_to_latex(F)
    return "";  // Nothing useful to add
end function;

// Filter uses format-specific data to reject or accept candidate numerical
// varieties.
function filter(X)
    // If a candidate X has an equation of the same weight as a var, reject.
    return #(SequenceToSet(EquationDegrees(X)) meet
                            SequenceToSet(Weights(X))) eq 0;
end function;

/////////////////////////////////////////////////////////////////////////
// Creation intrinsic
/////////////////////////////////////////////////////////////////////////

intrinsic FormatCompleteIntersection(D::[RngIntElt]) -> Fmt
{A codimension n format as the complete intersection of n polynomials of degrees d1..dn, where D is the sequence of integers [d1..dn]}
    // Sanity checks
    require #D ge 1: "The argument sequence must not be empty";
    require &and[ IsCoercible(Integers(),d) : d in D ]:
        "The argument sequence can only contain integers";
    ChangeUniverse(~D, Integers());
    require &and [ d gt 0 : d in D ]:
        "The integers in the argument sequence should all be strictly positive";

    // Calculate the numerical data of Koszul complex and build the format
    N := #D;
    betti_data := [ [Integers() ! 0] ];
    for i in [1..N] do
        Append(~betti_data, [ &+[D[i] : i in Q] : Q in Subsets({1..N},i) ]);
        // THINK: Al: is there mileage in caching such binomial sequences?
    end for;
    key_var_wts:=D;

    R:=PolynomialRing(Rationals(),#key_var_wts);
    eqns:=[ R.i : i in [1..N] ];
    // Create the basic format object
    F:=InternalCreateFormat( "complete intersection", #D, betti_data :
	    key_equations:=eqns,
	    key_weights := key_var_wts,
        extra_to_string := extra_to_string,
        extra_to_latex := extra_to_latex,
        filter := filter );
    F`numerical_data := D;
    // Return the format
    return F;
end intrinsic


//////////////////////////////////////////////////////////////////////
//                      Main search loop
//
// d = dimension of desired varieties.
// k = canonical degree - i.e. K_X = O_X(k).
//
// We want to search through a countable list which is ordered like this:
// (1) there is an integral paramater (called 'parameter' in the code)
// that starts at 'start' and increases in steps of one, either to
// 'finish' or never to finish.
// (2) for each value of the parameter, a 'ListFormats' function returns
// all possible formats having that value.
//
//////////////////////////////////////////////////////////////////////

intrinsic FormatSearchCompleteIntersection( d, k, c::RngIntElt :
	start := 0,
	finish := 20,
	minweight := 1,
	singtype := "isolated,quotient,canonical",
	file := false,
	latex := false,
	filterF := false,
	filterX := false ) -> SeqEnum
{Search for candidate varieties of dimension d, canonical class O(k) and codimension c in weighted projective space as complete intersections.}
    // Sanity check
    require c gt 0: "Argument 3 must be a strictly positive integer";
    // We incorporate list_formats here as a clean way to use codim 'c'.
    // It returns a sequence of c.i. formats of codim c and adj num k.
    list_formats:=function(k : filterF:=false)
        poss_F:=[];
        for D in Sort(Setseq(knapsack_with_repeats([2..k - 2*(c - 1)],k,c))) do
            F:=FormatCompleteIntersection(D);
            if filterF cmpeq false or filterF(F) then
                Append(~poss_F,F);
            end if;
        end for;
        return poss_F;
    end function;
    // Pass the data along to the generic search function
	bool,result:=search_format(d,k,list_formats :
		start := start,
		finish := finish,
		minweight := minweight,
		singtype := singtype,
		file := file,
		latex := latex,
		filterF := filterF,
		filterX := filterX );
	require bool: result;
	return result;
end intrinsic;
