//freeze;

/////////////////////////////////////////////////////////////////////////
// codim4slice§.m
/////////////////////////////////////////////////////////////////////////
// Codimension 4, hypersurface inside maximal pfaffians of skew 5x5 matrix.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////

import "utilities.m": search_format;
import "codim3.m": numerical, ijtoindex;
import "../Engine/strings.m": intseq_to_str, strict_upper_triangular_matrix_to_string;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// THINK: adapt to format
// Additional data about the format F as a string.
function extra_to_string(F,l)
    A:=F`numerical_data[2];     // a 5x5 matrix of weights
    w:=F`numerical_data[1];
    // Add the weights -- Are we in the integer or half-integer case?
    if Universe(w) cmpeq Integers() then
        s:="w: (" cat intseq_to_str(w) cat ")\n";
    else
        twow:=[Integers() | 2 * a : a in w];
        s:="w: 1/2(" cat intseq_to_str(twow) cat ")\n";
    end if;
    // Add the syzygy matrix
    s cat:= "Syzygy matrix weights: " cat
                strict_upper_triangular_matrix_to_string(A : brackets:="",
                    prefix:="                       ") cat "\n";
    return s;
end function;

// THINK: adapt to format
// Additional data about the format F as a LaTeX formatted string.
function extra_to_latex(F)
    A:=F`numerical_data[2];     // a 5x5 matrix of weights
    w:=F`numerical_data[1];
    // Add the weights -- Are we in the integer or half-integer case?
    if Universe(w) cmpeq Integers() then
        s:="$(" cat intseq_to_str(w) cat ")$&";
    else
        twow:=[Integers() | 2 * a : a in w];
        s := "$\\frac{1}{2}(" cat intseq_to_str(twow) cat ")$&";
    end if;
    // Add the syzygy matrix
    s cat:= strict_upper_triangular_matrix_to_string(A : brackets:="",latex:=true);
    return s;
end function;

// Filter uses format-specific data to reject or accept candidate numerical
// varieties.
function filter(X)
    // Reject this candidate if the c.i. equation eliminates a variable or
    // involves <= 2 vars (so is necessarily reducible).
    W := Weights(X);
    eqdegs := EquationDegrees(X);
    d := eqdegs[#eqdegs];
    if d in Weights(X) or #[ w : w in W | w lt d ] le 2 then
        return false;
    end if;
    // Reject this candidate if a Pfaffian equation eliminates a variable.
    F := Format(X);
    Fmx := F`numerical_data[2];
    if Fmx[1,2] eq 0 and
        &or[ Fmx[I[1],I[2]] in W : I in [[3,4],[3,5],[4,5]] ] then
        // Generic values in the matrix will eliminate a variable, so X is
        // really in lower codimension - not in this format.
        return false;
    end if;
    // THINK: more subtle rejection still needed: negative entry in matrix
    // where local tangent equation needed.
    return true;
end function;

/////////////////////////////////////////////////////////////////////////
// Creation intrinsic
/////////////////////////////////////////////////////////////////////////

intrinsic FormatCodimension4PfaffianSlice(w::SeqEnum,d::RngIntElt) -> Fmt
{A codimension 4 format as a hypersurface of degree d inside the five maximal pfaffians of a skew 5x5 matrix weighted by wi+wj, the integral or half-integral entries of the sequence w of length 5}
    // Sanity check
    require #w eq 5:
        "Argument 1 should be of length 5 and consist either entirely of integers or entirely of half-integers.";
    bool,neww:=CanChangeUniverse(w,Integers());
    if not bool then
        bool,neww:=CanChangeUniverse(w,Rationals());
        if bool then bool:=&and[IsHalfInteger(a) : a in w]; end if;
    end if;
    require bool:
        "Argument 1 should be of length 5 and consist either entirely of integers or entirely of half-integers.";
    w:=neww;
    require d gt 0: "Argument 2 should be a positive integer.";

    // Calculate the numerical data and equations
    D,E,k,A:=numerical(w);
    // We don't allow equations or syzygies of degree <= 0.
    require &and[ a gt 0 : a in D cat E cat [k] ]:
        "All equation and syzygy weights must be strictly positive";
    // Amend betti data as (codim 3 num) x (1 - t^d):
    k +:= d;
    E2 := [ a + d : a in D ] cat [ k - a - d : a in D ];
    D cat:= [d];
    E := [ k - a : a in D ];
    betti_data:=[PowerSequence(Integers()) | [0],D,E2,E,[k]];
    key_var_wts:=[Integers() | w[i] + w[j] : j in [i + 1..5],i in [1..4]]
        cat [d];
	// Since wts could be <= 0, we can't use a graded ring
    R:=PolynomialRing(Rationals(),#key_var_wts);
    matrix:=ZeroMatrix(R,5,5);
    for i in [1..4] do
        for j in [i + 1..5] do
            idx:=ijtoindex(i,j);
    	    matrix[i,j]:=R.idx;
            matrix[j,i]:=-R.idx;
        end for;
    end for;
    eqns:=Pfaffians(matrix,4) cat [R.11];
    // Create the basic format object
    F:=InternalCreateFormat( "wGrass(2,5) \\cap H", 4, betti_data :
	    key_equations:=eqns,
	    key_weights := key_var_wts,
        extra_to_string := extra_to_string,
        extra_to_latex := extra_to_latex,
        filter := filter );
    F`numerical_data:=<w,A>;
    // Return the format
    return F;
end intrinsic


//////////////////////////////////////////////////////////////////////
//                      Main search loop
//
// d = dimension of desired varieties.
// k = canonical degree - i.e. K_X = O_X(k).
//
// We want to search through a countable list which is ordered like this:
// (1) there is an integral paramater (called 'parameter' in the code)
// that starts at 'start' and increases in steps of one, either to
// 'finish' or never to finish.
// (2) for each value of the parameter, a 'ListFormats' function returns
// all possible formats having that value.
//
//////////////////////////////////////////////////////////////////////

// Return formats with adjunction degree k (i.e. -K_F = O(-k)).
function list_formats(k : filterF:=false)
    poss_F := [];
    for d in [2..Floor(k-3)] do
        target := (k-d)/2;
        w5range := [ target - 2 - j : j in [0..Ceiling(target) - 3] ]; // w5 > 0
        for w5 in w5range do
            w4range := [ w5 - j : j in [0..Ceiling(w5) - 1] ]; // w5 >= w4 > 0
            for w4 in w4range do
                 w3range := [ w4 - j : j in [0..Ceiling(w4) - 1] ]; // w4>=w3>0
                 for w3 in w3range do // w2+w3 >= 1
                    w2range := [ w3 - j : j in [0..Integers()!(2*w3 - 1)] ];
                    for w2 in w2range do
                        w1 := target - w5 - w4 - w3 - w2;
                        if w1 le w2 and w1 + w3 gt 0 then
                            w := [w1,w2,w3,w4,w5];
                            F := FormatCodimension4PfaffianSlice(w,d);
                            if filterF cmpeq false or filterF(F) then
                                Append(~poss_F,F);
                            end if;
                        end if;
                    end for;
                end for;
            end for;
        end for;
    end for;
    return poss_F;
end function;

intrinsic FormatSearchCodimension4PfaffianSlice( d, k::RngIntElt :
	start := 0,
	finish := 20,
	minweight := 1,
	singtype := "isolated,quotient,canonical",
	file := false,
	latex := false,
	filterF := false,
	filterX := false ) -> SeqEnum
{Search for candidate varieties of dimension d and canonical class O(k) in weighted projective space.}

    // Sanity check on our parameters we need to use before search_format.
    require Type(singtype) eq MonStgElt:
        "Parameter 'singtype' must be a string";
    require Type(minweight) eq RngIntElt:
        "Parameter 'minweight' must be an integer";
    require filterF cmpeq false or ISA(Type(filterF),Program):
        "Parameter 'filterF' must be a function";

    // Build a filter function based on our format-specific parameters.
    // First take into account any negative-weight matrix entry.
    // The point is this: if dim >= 4 and the 1,2 entry of the matrix is <= 0
    // then necessarily any candidate X has nonquotient singularity
    // which is worse than hyperquotient (emb dim > 1).
    if (d ge 4 and "quotient" in Split(singtype,",")) then
        f1:=function(F)
            if ((F`numerical_data)[2])[1,2] le 0 then
                return false;
            end if;
            return filterF cmpeq false or filterF(F);
        end function;
    else
        f1:=filterF;
    end if;
    // Since the matrix degrees increase, if wt(M[1,3]) < minweight, then
    // the first Pfaffian is identically zero or factorises, so reject format.
    if minweight gt 1 then
        f2 := function(F)
            if (F`numerical_data[2])[1,3] lt minweight then
                return false;
            end if;
            return f1 cmpeq false or f1(F);
        end function;
    else
        f2:=f1;
    end if;

	bool,result:=search_format(d,k,list_formats :
		start := start,
		finish := finish,
		minweight := minweight,
		singtype := singtype,
		file := file,
		latex := latex,
		filterF := f2,
		filterX := filterX );
	require bool: result;
	return result;
end intrinsic;
