//////////////////////////////////////////////////////////////////////
//      Format: codimension 5 orthogonal grassmannian
//
// See [Corti--Reid] for the maths.
//
// This is the only file that understands this format so any
// functions needing knowledge of the format must live here.
//
// This file must have a creation intrinsic:
//      FormatXXX       where XXX = format name
// and a search intrinsic
//      FormatSearchXXX
//
// The following intrinsics are written at a generic level. If there
// is a better format-specific version, then it can be written here with
// the same name prefixed by "codim5ograss" and will be called automatically:
//      AnalyseZeroStratum      (generic version does nothing)
//      IsOnFormat      (generic version uses the equations of the key variety)
//      FirstGuessOfWeightsFromEquations
//      MaximumWeightNonCone    (generic version returns the max generic wt)
//      ExtraFormatToString     (generic version prints nothing extra)
//      CandidateWeights
//
//////////////////////////////////////////////////////////////////////

import "utilities.m": search_format;
import "../Engine/utilities.m": knapsack_with_repeats;
import "../Engine/strings.m": strict_upper_triangular_matrix_to_string,
    num_to_latex;

//////////////////////////////////////////////////////////////////////
//			Local functions
//////////////////////////////////////////////////////////////////////

function ijtoindex(i,j)
        return Index(
                [[1,2],[1,3],[1,4],[1,5],[2,3],[2,4],[2,5],[3,4],[3,5],[4,5]],
                        [Integers()|i,j]);
end function;

function extra_to_string(F)
	return "";		// THINK: anything to add here?
end function;

function extra_to_latex(F)
	return "";		// THINK: anything to add here?
end function;

/* THINK: reinstate this for export version
function extra_to_latex(F)
        u := F`numerical_data[4];
        w := F`numerical_data[5];
        s := "\\begin{array}{c}" cat num_to_latex(u) cat "\\\\";
        if IsIntegral(w[1]) then
                s cat:= ToString(w : brackets := "(", latex := true)
                        cat "\\end{array}&";
        else
                twow := [Integers()| 2*a : a in w];
                s cat:= "\\frac{1}{2}" cat
                        ToString(twow : brackets := "(", latex := true)
                        cat "\\end{array}&";
        end if;
        x := F`numerical_data[1];
        nx := F`numerical_data[2];
        A := F`numerical_data[3];
        s cat:= ToString(x : latex := true) cat "&"
                cat ToString(nx : latex := true) cat "&"
                cat strict_upper_triangular_matrix_to_string(A : brackets := "",
                                                                latex := true);
        return s;
end function;
*/

//////////////////////////////////////////////////////////////////////
//			Creation
//////////////////////////////////////////////////////////////////////

intrinsic FormatCodimension5OrthogonalGrassmannian(u::RngIntElt,w::SeqEnum) -> Fmt
{A codimension 5 format as the orthogonal grassmannian determined by the
integer u and length 5 vector w of integers or half-integers}
	// THINK: sanity checks on w to do

	// Calculate the numerical data and equations
	s := &+ w;
	d := s + 2*u;
	eqns := [ Integers() | d - wi : wi in w ] cat [ Integers() | d + wi : wi in w ];
// THINK: do I really need all equations to be in positive degree?
	require &and [ e gt 0 : e in eqns ]:
		"Every equation in a format must have strictly positive weight";
	syz1 := [ Integers() |2*d-u] cat [Integers() |  2*d - u - s + wi : wi in w ]
		cat [Integers() |  2*d -u - w[i] - w[j] : j in [i+1..5], i in [1..4] ];
	syz2 := [ Integers() |2*d+u] cat [Integers() |  2*d + u + s - wi : wi in w ]
		cat [Integers() |  2*d + u + w[i] + w[j] : j in [i+1..5], i in [1..4] ];
	syz3 := [ Integers() | 3*d - wi : wi in w ] cat [Integers() |  3*d + wi : wi in w ];
	k :=  Integers() ! (4*d);
        betti_data := [ [Integers()|0],eqns,syz1,syz2,syz3,[k]];
	Aplusu := Matrix(5,[Integers()|u+wi+wj: wi,wj in w]);
        wplusu := [Integers()| u + s - wi : wi in w ];
	all_wts := [Integers()|u] cat [Integers()|Aplusu[i,j]: j in [i+1..5],i in [1..4]] cat wplusu;
        R<x,x12,x13,x14,x15,x23,x24,x25,x34,x35,x45,x1,x2,x3,x4,x5>
		:= PolynomialRing(Rationals(),#all_wts);
	x := Matrix(1,1,[R.1]);
	matrix := Matrix(5,5,[R|0:i in [1..25]]);
        for i in [1..4] do
                for j in [i+1..5] do
                        matrix[i,j] := R.(ijtoindex(i,j)+1);
                        matrix[j,i] := -R.(ijtoindex(i,j)+1);
                end for;
        end for;
	vector := Matrix(5,1,[R.i:i in [12..16]]);
	pfaff_eqs := Eltseq(vector * x - Matrix(5,1,Pfaffians(matrix,4)));
	product_eqs := Eltseq(matrix * vector);
	// Build the format and return
	F := InternalCreateFormat( "codim 5 OGrass(5,10)", 5, betti_data :
		key_equations := pfaff_eqs cat product_eqs,
		key_weights := all_wts,
		extra_to_string := extra_to_string,
		extra_to_latex := extra_to_latex );
	F`numerical_data := <u,wplusu,Aplusu,u,w>;
        return F;
end intrinsic


//////////////////////////////////////////////////////////////////////
//			Main search loop
//////////////////////////////////////////////////////////////////////

// For integer n, return all cod 5 orthog grass formats F(u,w) with 4u+2|w|=n.
// This works because our ordering forces w1 + w2 + u to be the next
// smallest weight.
forward even_w, odd_w;
// THINK: how does min_wt get through?
// function ListFormatsCodimension5OrthogonalGrassmannian(n,min_wt)
function list_formats(n: filterF:=false)
    min_wt := 1;    // THINK
    formats := [];
    u := min_wt;
    repeat
	w_sum := (n - 4*u) / 2;	// parameter n = 4*u + 2*w_sum.
	if IsEven(n) then
		poss_w := even_w(u,w_sum,min_wt);
	else
		poss_w := odd_w(u,w_sum,min_wt);
	end if;
    	formats cat:= [<u,w> : w in poss_w];
	u +:= 1;
    until u gt n/4;
    return [FormatCodimension5OrthogonalGrassmannian(F[1],F[2]):F in formats];
end function;

intrinsic FormatSearchCodimension5OrthogonalGrassmannian( d, k::RngIntElt :
    start := 0,
    finish := 20,
    minweight := 1,
    singtype := "isolated,quotient,canonical",
    file := false,
    latex := false,
    filterF := false,
    filterX := false ) -> SeqEnum
{Search for numerical varieties of dimension d and canonical class O(k)
in weighted projective space.}
    bool,result:=search_format(d,k,list_formats :
        start := start,
        finish := finish,
        minweight := minweight,
        singtype := singtype,
        file := file,
        latex := latex,
        filterF := filterF,
        filterX := filterX );
    require bool: result;
    return result;
end intrinsic;

/*
    bool,min_wt := IsCoercible(Integers(),minweight);
    require bool: "Parameter 'minweight' must be an integer";
    require IsCoercible(Integers(),start): "parameter start must be an integer";
    results := [];

    file := file eq "" select false else Open(file,"w");
    if Type(file) ne BoolElt then
        HEADER := "Candidate varieties X of dimension" cat
            IntegerToString(d) cat " in wps with \n" cat
            "\tK_X = O(" cat IntegerToString(k) cat ")\n" cat
            "\tcodim = 5\n" cat
            "\tsings = " cat singtype cat"\n\n";
        Put(file,HEADER);
        Flush(file);
    end if;
    latexfile := latexfilename eq "" select false else Open(latexfilename,"w");

    parameter := start;
    repeat
        vprint Format: "------------------",parameter;
	old_number := #results;
        tt := Cputime();
        formats := ListFormatsCodimension5OrthogonalGrassmannian(parameter,minweight);
        for data in formats do
	    F := FormatCodimension5OrthogonalGrassmannian(data[1],data[2]);
            Xs := CandidatesFromFormat(F,d,k :
			singtype:=singtype, minweight:=min_wt);
	    results cat:= Xs;
            if Type(file) ne BoolElt then
                for X in Xs do
                    Put(file,Sprintf("%o\n",X));
                end for;
            end if;
        end for;
// THINK: save to file if file is set
        parameter +:= 1;
        if Type(file) ne BoolElt then
            Flush(file);
        end if;
        vprintf Format: "New: %o\tTotal: %o\n",#results-old_number,#results;
        vprint Format: "Time:",Cputime(tt);
    until parameter gt finish;
    return results;
end intrinsic;
*/

// THINK: We're assuming that w2<=...<=w4. Is this OK?
function even_w(u,w_sum,min_wt)
    w_sum := Integers() ! w_sum;
    solns := {};
    w1 := Ceiling(4/3*(min_wt - u-1/4*w_sum));
    repeat
        soln0 := knapsack_with_repeats([AbsoluteValue(w1)..(w_sum-w1)],w_sum-w1,4);
        solns join:= { [w1] cat Reversion(x) : x in soln0 };
        w1 +:= 1;
    until (w1 gt w_sum/5);
    return Sort(Setseq(solns));
end function;

// THINK: Think carefully about the ceiling stuff here -- is it right?
function odd_w(u,w_sum,min_wt)
    // w_sum is a half-integer.
    min_wt_2 := Integers() ! (2 * min_wt);
    solns := {};
    w1_2 := Ceiling(8/3*(min_wt - u-1/4*w_sum));
    if IsEven(w1_2) then
	w1_2 +:= 1;	// the resulting w1 = w1_2/2 must be a half integer.
    end if;
    repeat
	max_w_2 := Integers() ! (2*w_sum-w1_2);
        soln0 := knapsack_with_repeats([AbsoluteValue(w1_2)..max_w_2 by 2],
			 Integers() ! (2*w_sum-w1_2),4);
        solns join:= { [w1_2/2] cat [a/2 : a in Reversion(x)] : x in soln0 };
        w1_2 +:= 2;
    until (w1_2 gt 2*w_sum/5);
    return Sort(Setseq(solns));
end function;


//////////////////////////////////////////////////////////////////////
//         	Optional Intrinsics
// THINK: the two big ones here are commented out - generic version takes over.
//////////////////////////////////////////////////////////////////////


/*
intrinsic AnalyseZeroStratum(F::GrphRes,W::[RngIntElt],r::RngIntElt,k::RngIntElt        : singtype := "terminal") -> BoolElt,SetEnum
{True if there is no contradiction to the unique ambient index r point being terminal (or absent) on a realisation of F. Also returns a set of singularities of
index r that could appear on a numerical variety with canonical class O(k) in fo
rmat F embedded in wps P(W). Assumes that there is a unique point in P(W) of ind
ex divisible by r}
    require F`formatname eq "codim5ograss": "F is in the wrong format";
    require #[w:w in W|IsDivisibleBy(w,r)] eq 1:
    		"W must have exactly one element not coprime to r";
    set_of_sings := {};
    ambient_wts := W;
    local_wts := W;
    // First, if this point does not lie on the general X (i.e. if one
    // of the eqn degs is divisible by r), then that's fine - return no sings.
    if not IsOnFormat(F,r) then
        return true, {};
    end if;
    // find the weights of polarising variables after eliminating
    // vars using some (or all) of the equations.

    // If y is the var of wt r, we search all terms of all equations
    // to see whether a term of the form y^N*z could appear for some var z.
    // If so, throw away the weight of z from our list and also throw
    // away the eqn we used to eliminate z from our list of available eqns.

    local_wts_copy := Exclude(local_wts,r);
    u := F`numerical_data[1];
    w := F`numerical_data[2];
    A := F`numerical_data[3];

    // check the 10 generic variables appearing in the matrix first.
    // e.g. x12 mults the 3 complementary matrix entries x34, x35, x45
    // and the two vector entries x1, x2.
    // We only look for elimination inside these vars.
    locations_in_A := [ [i,j] : j in [i+1..5], i in [1..4] |
                    IsDivisibleBy(A[i,j],r) ];
    for pivot in locations_in_A do
	    i := pivot[1];
	    j := pivot[2];
            local_wts := local_wts_copy;
            tangent_eqns123 := ComplementaryWeights(A,i,j);
	    tangent_eqns45 := [ w[i], w[j] ];
	    local_eqn_wts := [ Integers() | ];
            for d in tangent_eqns123 cat tangent_eqns45 do
                for a in local_wts do
                    residual := d-a;
                    if residual ge 0 and residual mod r eq 0 then
                        // can eliminate the var of wt a using the eqn of wt d near P_r
                        Exclude(~local_wts,a);
                        continue d;     // we've used this equation 'd' to eliminate var 'a' so move to next equation.
                    end if;
                end for;
                Append(~local_eqn_wts,d);
            end for;
            Include(~set_of_sings,
                Singularity(Dimension(F,ambient_wts),r,local_wts,local_eqn_wts,k));
    end for;

    // next check the 5 generic variables appearing in the vector.
    // e.g. x1 mults x and the 4 entries in row/col 1, x12,...,x15
    locations_in_w := [ i : i in [1..5] | w[i] mod r eq 0 ];
    for pivot in locations_in_w do
            local_wts := local_wts_copy;
	    tangent_eqns1 := [ u ];
            tangent_eqns2345 := [ A[pivot,j] : j in [1..5] | j ne pivot ];
	    local_eqn_wts := [ Integers() | ];
            for d in tangent_eqns1 cat tangent_eqns2345 do
                for a in local_wts do
                    residual := d-a;
                    if residual ge 0 and residual mod r eq 0 then
                        // can eliminate the var of wt a using the eqn of wt d near P_r
                        Exclude(~local_wts,a);
                        continue d;     // we've used this equation 'd' to eliminate var 'a' so move to next equation.
                    end if;
                end for;
                Append(~local_eqn_wts,d);
            end for;
            Include(~set_of_sings,
                Singularity(Dimension(F,ambient_wts),r,local_wts,local_eqn_wts,k));
    end for;

    // finally check generic variable x; it mults x1..x5, the vector vars.
    local_wts := local_wts_copy;
    tangent_eqns12345 := w;
    if u mod r eq 0 then
	    local_eqn_wts := [ Integers() | ];
	    for d in tangent_eqns12345 do
		for a in local_wts do
		    residual := d-a;
		    if residual ge 0 and residual mod r eq 0 then
			// can eliminate the var of wt a using the eqn of wt d near P_r
			Exclude(~local_wts,a);
			continue d;     // we've used this equation 'd' to eliminate var 'a' so move to next equation.
		    end if;
		end for;
		Append(~local_eqn_wts,d);
	    end for;
	    Include(~set_of_sings,
		Singularity(Dimension(F,ambient_wts),r,local_wts,local_eqn_wts,k));
    end if;

    set_of_sings := { P : P in set_of_sings | IsOfType(P,singtype) };
    return #set_of_sings ne 0, set_of_sings;
end intrinsic;
*/

/*
intrinsic IsOnFormat(F::GrphRes,a::RngIntElt) -> BoolElt
{True if an ambient 0-stratum 1/a necessarily lies on any model of F}
// i.e.  true iff there are complementary entries of syz mx A that are 0 mod a
// or x and some xi are both 0 mod a (so the pfaff eqns may not vanish at 1/a)
// OR some xi and an xij (or xji) are both 0 mod a (so a product eqn is not 0).
  u := F`numerical_data[1];
  w := F`numerical_data[2];
  A := F`numerical_data[3];
  // check the pfaffian equations
  if u mod a eq 0 and &or[ b mod a eq 0 : b in w ] then
  	return false;
  end if;
  locations_in_A := [ [i,j] : j in [i+1..5], i in [1..4] |
                IsDivisibleBy(A[i,j],a) ];
  for i in [1..#locations_in_A] do
    for j in [i+1..#locations_in_A] do
      entry1 := locations_in_A[i];
      entry2 := locations_in_A[j];
      if #{entry1[1],entry1[2],entry2[1],entry2[2]} eq 4 then
        return false;
      end if;
    end for;
  end for;
  // check the product equations
  for i in [1..#w] do
	b := w[i];
  	if b mod a eq 0 then
		for j in [c : c in [1..5] | c ne i ] do
			if A[i,j] mod a eq 0 then
				return false;
			end if;
		end for;
	end if;
  end for;
  return true;
end intrinsic;
*/

/*
forward extra_to_latex;
intrinsic codim5ograssExtraFormatToString(F::GrphRes : latex := false) -> MonStgElt
{Additional format-specific data about the numerical formal F as a string}
        // We pass the LaTeX case off to a seperate helper function
        if latex then
                return extra_to_latex(F);
        end if;
        // Handle the non-LateX case here
        u := F`numerical_data[4];
        w := F`numerical_data[5];
        s := "u:\t" cat ToString(u);
        if IsIntegral(w[1]) then
                s cat:= "\nw:\t" cat ToString(w : brackets:="(") cat "\n";
        else
                twow := [Integers()| 2*a : a in w];
                s cat:= "\nw:\t1/2" cat ToString(twow : brackets:="(") cat "\n";
        end if;
        x := F`numerical_data[1];
        nx := F`numerical_data[2];
        A := F`numerical_data[3];
        s cat:= "x weight:\t" cat ToString(x) cat "\nx_i weights:\t"
            cat ToString(nx : brackets:="")
            cat "\nx_ij weights:\t" cat
            strict_upper_triangular_matrix_to_string(A:brackets:="",prefix:="\t\t")
            cat "\n";
        return s;
end intrinsic;
*/

